# Licensed under the GPLv3 - see LICENSE
"""The VDIFFileReaderInfo property.

Includes information about threads and frame sets.
"""
from ..base.file_info import FileReaderInfo, info_item


__all__ = ['VDIFFileReaderInfo']


class VDIFFileReaderInfo(FileReaderInfo):
    attr_names = (('format', 'edv', 'number_of_frames', 'thread_ids',
                   'number_of_framesets')
                  + FileReaderInfo.attr_names[2:])
    """Attributes that the container provides."""

    edv = info_item(needs='header0', doc=(
        'Extended data version (False if legacy).'))

    @info_item
    def thread_ids(self):
        # To get the thread_ids and thus the real sample shape,
        # need to check frame sets.
        with self._parent.temporary_offset(0) as fh:
            return fh.get_thread_ids()

    @info_item
    def header0(self):
        with self._parent.temporary_offset(0) as fh:
            # Almost all bytes are interpretable as headers,
            # so we need a basic sanity check.
            return fh.find_header(maximum=0)

    # Some headers also need frame rate, but fine to let that
    # lead to an error.
    @info_item(needs='header0')
    def start_time(self):
        return self.header0.get_time(frame_rate=self.frame_rate)

    @info_item(needs=('header0', 'thread_ids'))
    def sample_shape(self):
        return (len(self.thread_ids), self.header0.nchan)

    @info_item(needs=('number_of_frames', 'thread_ids'))
    def number_of_framesets(self):
        number_of_framesets = self.number_of_frames / len(self.thread_ids)
        if number_of_framesets % 1 == 0:
            return int(number_of_framesets)

        else:
            self.warnings['number_of_framesets'] = (
                'file contains non-integer number ({}) of '
                'framesets'.format(number_of_framesets))
            return None



# Licensed under the GPLv3 - see LICENSE
"""
Base definitions for baseband Headers, in particular for the VLBI types.

Defines a number of helpers to construct a Header class that hold header
words corresponding to a frame header, and provides access to the values
encoded in those words via a dict-like interface.  Definitions for headers
are constructed using a HeaderParser class (which is targeted specifically
at the VLBI formats).
"""
import sys
import struct
import warnings
import functools
from copy import copy

import numpy as np
from astropy.utils import sharedmethod

from .utils import fixedvalue


__all__ = ['four_word_struct', 'eight_word_struct',
           'make_parser', 'make_setter', 'get_default',
           'ParserDict', 'HeaderParserBase', 'HeaderParser',
           'ParsedHeaderBase', 'VLBIHeaderBase']


four_word_struct = struct.Struct('<4I')
"""Struct instance that packs/unpacks 4 unsigned 32-bit integers."""
eight_word_struct = struct.Struct('<8I')
"""Struct instance that packs/unpacks 8 unsigned 32-bit integers."""


def make_parser(word_index, bit_index, bit_length, default=None):
    """Construct a function that converts specific bits from a header.

    The function acts on a tuple/array of 32-bit words, extracting given bits
    from a specific word and convert them to bool (for single bit) or integer.

    The parameters are those that define header keywords, and all parsers do
    ``(words[word_index] >> bit_index) & ((1 << bit_length) - 1)``, except that
    that they have been optimized for the specific cases of single bits,
    full words, and items starting at bit 0.  As a special case, bit_length=64
    allows one to extract two words as a single (long) integer.

    Parameters
    ----------
    word_index : int
        Index into the tuple of words passed to the function.
    bit_index : int
        Index to the starting bit of the part to be extracted.
    bit_length : int
        Number of bits to be extracted.

    Returns
    -------
    parser : function
        To be used as ``parser(words)``.
    """
    if bit_length == 1:
        def parser(words):
            return (words[word_index] & (1 << bit_index)) != 0

    elif bit_length == 32:
        assert bit_index == 0

        def parser(words):
            return words[word_index]

    elif bit_length == 64:
        assert bit_index == 0

        def parser(words):
            return words[word_index] + (words[word_index + 1] << 32)

    else:
        bit_mask = (1 << bit_length) - 1  # e.g., bit_length=8 -> 0xff
        if bit_index == 0:
            def parser(words):
                return words[word_index] & bit_mask

        else:
            def parser(words):
                return (words[word_index] >> bit_index) & bit_mask

    return parser


def make_setter(word_index, bit_index, bit_length, default=None):
    """Construct a function that uses a value to set specific bits in a header.

    The function will act on a tuple/array of words, setting given bits
    from a given word using a value.

    The parameters are just those that define header keywords.

    Parameters
    ----------
    word_index : int
        Index into the tuple of words passed to the function.
    bit_index : int
        Index to the starting bit of the part to be extracted.
    bit_length : int
        Number of bits to be extracted.
    default : int or bool or None
        Possible default value to use in function if no default is passed on.

    Returns
    -------
    setter : function
        To be used as ``setter(words, value)``.
    """
    def setter(words, value):
        bit_mask = (1 << bit_length) - 1
        if value is None:
            if default is None:
                raise ValueError("no default value so cannot set to 'None'.")
            value = default
        elif value is True:
            value = bit_mask
        elif np.any(value & bit_mask != value):
            raise ValueError("{0} cannot be represented with {1} bits"
                             .format(value, bit_length))
        if bit_length == 64:
            word1 = value & (1 << 32) - 1
            word2 = value >> 32
            words[word_index] = word1
            words[word_index + 1] = word2
        else:
            word = words[word_index]
            # Zero the part to be set, and add the value.
            bit_mask <<= bit_index
            word = ((word | bit_mask) ^ bit_mask) | (value << bit_index)
            words[word_index] = word
        return words

    return setter


def get_default(word_index, bit_index, bit_length, default=None):
    """Return the default value from a header keyword.

    Since it is called with the full description, it just returns
    the last item, defaulted to `None`.
    """
    return default


class ParserDict:
    """Create a lazily evaluated dictionary of parsers, setters, or defaults.

    Implemented as a non-data descriptor.  When first called on an instance,
    it will create a dict under the name of itself in the instance's
    ``__dict__``, which means that any further attribute access will return
    that dict instead of this descriptor.

    Parameters
    ----------
    function : callable
        Function that can be used to create a parser or setter, or get the
        default, based on a header keyword description.  Typically one of
        ``make_parser``, ``make_setter``, or ``get_default``.

    """

    def __init__(self, function):
        self.function = function

    def __set_name__(self, owner, name):
        self.name = name
        self.__doc__ = f"Lazily evaluated dict of {name}"

    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        # Create dict of functions/defaults.
        d = {key: self.function(*definition)
             for key, definition in instance.items()}
        # Override ourselves on the instance.
        setattr(instance, self.name, d)
        return d

    def __repr__(self):
        return f"{self.__class__.__name__}({self.function}"


class HeaderParserBase(dict):
    """Parser & setter for header keywords.

    A dictionary of header keywords, with values that describe how they are
    encoded in a given header.  Initialisation is as a normal dict,
    with (ordered) key, value pairs, with each value a tuple containing
    information that describes how the value is encoded, and any default.

    The actual implementation is done by instances of
    `~baseband.base.header.ParserDict` called ``parsers``, ``setters``, and
    ``defaults``, which return functions that get a given keyword from
    header words, set the corresponding part of the header words to a value,
    or return the default value (if defined).  To speed up access to those,
    they are precalculated on first access rather than calculated on the fly.

    """
    def copy(self):
        return self.__class__(self)

    def __or__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented

        if sys.version_info >= (3, 9):  # pragma: no cover
            return self.__class__(super().__or__(other))

        result = self.__class__(self)
        result.update(other)
        return result

    # Backwards compatibility for code written for baseband < 4.0.
    __add__ = __or__

    def _clear_caches(self):
        """Clear the caches of the parser dicts. To be done on any change."""
        for key in set(self.__dict__):
            if isinstance(self.__class__.__dict__[key], ParserDict):
                del self.__dict__[key]


# Overwrite all dict methods that change the contents to clear the
# cache of the parsers and setters.
def make_wrapped_method(method):
    @functools.wraps(getattr(dict, method))
    def wrapped(self, *args, **kwargs):
        result = getattr(super(HeaderParserBase,
                               self), method)(*args, **kwargs)
        self._clear_caches()
        return result
    return wrapped


for method in ('__setitem__', 'update', 'pop', 'popitem', 'clear'):
    setattr(HeaderParserBase, method, make_wrapped_method(method))


class HeaderParser(HeaderParserBase):
    """Parser & setter for VLBI header keywords.

    A dictionary of header keywords, with values that describe how they are
    encoded in a given VLBI header.  Initialisation is as a normal dict,
    with (ordered) key, value pairs, with each value a tuple containing:

    word_index : int
        Index into the header words for this key.
    bit_index : int
        Index to the starting bit of the part used for this key.
    bit_length : int
        Number of bits.
    default : int or bool or None
        Possible default value to use in initialisation (e.g., a sync pattern).

    The class provides dict-like properties ``parsers``, ``setters``, and
    ``defaults``, which return functions that get a given keyword from header
    words, set the corresponding part of the header words to a value, or
    return the default value (if defined).  To speed up access to those,
    they are precalculated on first access rather than calculated on the fly.

    """
    parsers = ParserDict(make_parser)
    setters = ParserDict(make_setter)
    defaults = ParserDict(get_default)


class ParsedHeaderBase:
    """Base class for all baseband headers using parsers.

    Defines a number of common routines.

    Generally, the actual class should define:

      _header_parser : HeaderParser instance corresponding to this class.

      _properties : tuple of properties accessible/usable in initialisation

    It also should define properties that tell the size (getters *and*
    setters, or use a `baseband.base.utils.fixedvalue` if the
    value is the same for all instances):

      payload_nbytes : number of bytes used by payload

      frame_nbytes : total number of bytes for header + payload

      get_time, set_time, and a corresponding time property :
           time at start of payload

    Parameters
    ----------
    words : tuple or list
        Header words.  If given as a tuple, the header is immutable.
    verify : bool, optional
        Whether to do basic verification of integrity.

    """

    _properties = ('payload_nbytes', 'frame_nbytes', 'time')
    """Properties accessible/usable in initialisation for all headers."""

    def __init__(self, words, verify=True):
        self.words = words
        if verify:
            self.verify()

    def verify(self):
        """Base verification always passes.

        Only here for subclasses to be able to do super().
        """
        pass  # pragma: no cover

    def copy(self, **kwargs):
        """Create a mutable and independent copy of the header.

        Keyword arguments can be passed on as needed by possible subclasses.
        """
        kwargs.setdefault('verify', False)
        new = self.__class__(copy(self.words), **kwargs)
        new.mutable = True
        return new

    def __copy__(self):
        return self.copy()

    @property
    def mutable(self):
        """Whether the header can be modified."""
        if isinstance(self.words, tuple):
            return False
        word0 = self.words[0]
        try:
            self.words[0] = 0
        except Exception:
            return False
        else:
            self.words[0] = word0
            return True

    @mutable.setter
    def mutable(self, mutable):
        if isinstance(self.words, np.ndarray):
            self.words.flags['WRITEABLE'] = mutable
        elif isinstance(self.words, tuple):
            if mutable:
                self.words = list(self.words)
        elif isinstance(self.words, list):
            if not mutable:
                self.words = tuple(self.words)
        else:
            raise TypeError("do not know how to set mutability of '.words' "
                            "of class {0}".format(type(self.words)))

    @classmethod
    def fromvalues(cls, *args, **kwargs):
        """Initialise a header from parsed values.

        Here, the parsed values must be given as keyword arguments, i.e., for
        any ``header = cls(<words>)``, ``cls.fromvalues(**header) == header``.

        However, unlike for the `fromkeys` class method, data can also be set
        using arguments named after header methods, such as ``time``.

        Parameters
        ----------
        *args
            Possible arguments required to initialize an empty header.
        **kwargs
            Values used to initialize header keys or methods.
        """
        # Initialize an empty header, and update it with the keyword arguments.
        self = cls(None, *args, verify=False)
        # Set defaults in keyword arguments.
        for key in set(self.keys()).difference(kwargs.keys()):
            default = self._header_parser.defaults[key]
            if default is not None:
                kwargs[key] = default

        self.update(**kwargs)
        return self

    @classmethod
    def fromkeys(cls, *args, **kwargs):
        """Initialise a header from parsed values.

        Like fromvalues, but without any interpretation of keywords.

        Raises
        ------
        KeyError : if not all keys required are present in ``kwargs``
        """
        self = cls(None, *args, verify=False)
        not_in_both = (set(self.keys()).symmetric_difference(kwargs)
                       - {'verify'})
        if not_in_both:
            not_in_kwarg = set(self.keys()).difference(kwargs)
            not_in_self = set(kwargs).difference(self.keys()) - {'verify'}
            msg_parts = []
            for item, msg in ((not_in_kwarg, "is missing keywords ({0})"),
                              (not_in_self, "contains extra keywords ({0})")):
                if item:
                    msg_parts.append(msg.format(item))

            raise KeyError("input list " + " and ".join(msg_parts))

        self.update(**kwargs)
        return self

    def update(self, *, verify=True, **kwargs):
        """Update the header by setting keywords or properties.

        Here, any keywords matching header keys are applied first, and any
        remaining ones are used to set header properties, in the order set
        by the class (in ``_properties``).

        Parameters
        ----------
        verify : bool, optional
            If `True` (default), verify integrity after updating.
        **kwargs
            Arguments used to set keywords and properties.
        """
        # First use keywords which are also keys into self.
        for key in set(kwargs.keys()).intersection(self.keys()):
            self[key] = kwargs.pop(key)

        # Next, use remaining keyword arguments to set properties.
        # Order is important, so we cannot use an intersection as above.
        if kwargs:
            for key in self._properties:
                if key in kwargs:
                    setattr(self, key, kwargs.pop(key))

            if kwargs:
                warnings.warn("some keywords unused in header update: {0}"
                              .format(kwargs))

        if verify:
            self.verify()

    def __getitem__(self, item):
        """Get the value of a particular header item from the header words."""
        try:
            return self._header_parser.parsers[item](self.words)
        except KeyError:
            raise KeyError("{0} header does not contain {1}"
                           .format(self.__class__.__name__, item))

    def __setitem__(self, item, value):
        """Set the value of a particular header item in the header words.

        If value is `None`, set the item to its default value (if it exists);
        if `True`, set all bits in the item (i.e., set item to its maximum).
        """
        try:
            self._header_parser.setters[item](self.words, value)
        except KeyError:
            raise KeyError("{0} header does not contain {1}"
                           .format(self.__class__.__name__, item))
        except(TypeError, ValueError):
            if not self.mutable:
                raise TypeError("header is immutable. Set '.mutable` attribute"
                                " or make a copy.")
            else:
                raise

    def keys(self):
        """All keys defined for this header."""
        return self._header_parser.keys()

    def _ipython_key_completions_(self):
        # Enables tab-completion of header keys in IPython.
        return self.keys()

    def __contains__(self, key):
        return key in self.keys()

    def __eq__(self, other):
        return (type(self) is type(other)
                and np.all(np.array(self.words, copy=False)
                           == np.array(other.words, copy=False)))

    def _repr_value(self, key, value):
        return str(value)

    def __repr__(self):
        name = self.__class__.__name__
        outs = [f"{k}: {self._repr_value(k, self[k])}" for k in self.keys()]
        return "<{} {}>".format(name, (",\n  " + " "*len(name)).join(outs))


class VLBIHeaderBase(ParsedHeaderBase):
    """Base class for all VLBI headers.

    Defines a number of common routines.

    Generally, the actual class should define:

      _struct : `~struct.Struct` instance that can pack/unpack header words.

      _header_parser : `HeaderParser` instance corresponding to this class.

      _properties : tuple of properties accessible/usable in initialisation

      _invariants : set of keys of invariant header parts for a given type.

      _stream_invarants : set of keys of invariant header parts for a stream.

    It also should define properties that tell the size (getters *and*
    setters, or use a `baseband.base.utils.fixedvalue` if the
    value is the same for all instances):

      payload_nbytes : number of bytes used by payload

      frame_nbytes : total number of bytes for header + payload

      get_time, set_time, and a corresponding time property :
           time at start of payload

    Parameters
    ----------
    words : tuple or list of int, or None
        header words (generally, 32 bit unsigned int).  If given as a tuple,
        the header is immutable.  If `None`, set to a list of zeros for
        later initialisation (and skip any verification).
    verify : bool, optional
        Whether to do basic verification of integrity.  For the base class,
        checks that the number of words is consistent with the struct size.
    """

    # TODO: should [_stream]_invarants be defined through some subclass init??
    # TODO: perhaps from some hints in the headerparser definition?

    # Define a bare _struct to avoid sphinx complaints about nbytes.
    _struct = struct.Struct('')
    """Structure for the header words.  To be overridden by subclasses."""

    def __init__(self, words, verify=True):
        if words is None:
            words = [0] * (self._struct.size // 4)
            verify = False

        super().__init__(words, verify=verify)

    def verify(self):
        """Verify that the length of the words is consistent.

        Subclasses should override this to do more thorough checks.
        """
        assert len(self.words) == (self._struct.size // 4)

    @sharedmethod
    def invariants(self):
        """Set of keys of invariant header parts.

        On the class, this returns keys of parts that are shared by
        all headers for the type, on an instance, those that are
        shared with other headers in the same file.

        If neither are defined, returns 'sync_pattern' if the header
        containts that key.
        """

        if not isinstance(self, type) and hasattr(self, '_stream_invariants'):
            return self._stream_invariants

        elif hasattr(self, '_invariants'):
            return self._invariants

        elif 'sync_pattern' in getattr(self, '_header_parser', {}):
            return {'sync_pattern'}

        else:
            return set()

    @sharedmethod
    def invariant_pattern(self, invariants=None, **kwargs):
        """Pattern and mask shared between headers of a type or stream.

        This is mostly for use inside
        :meth:`~baseband.base.base.VLBIFileReaderBase.locate_frames`.

        Parameters
        ----------
        invariants : set of str, optional
            Set of keys to header parts that are shared between all headers
            of a given type or within a given stream/file.  Default: from
            `~baseband.base.header.VLBIHeaderBase.invariants()`.
        **kwargs
            Keyword arguments needed to instantiate an empty header.
            (Mostly for Mark 4).

        Returns
        -------
        pattern : list of int
            The pattern that is shared between headers. If called on
            an instance, just the header words; if called on a class,
            words with defaults for the relevant parts set.
        mask : list of int
            For each entry in ``pattern`` a bit mask with bits set for
            the parts that are invariant.
        """

        if invariants is None:
            invariants = self.invariants()

        if not invariants:
            raise ValueError("cannot create an invariant_mask without "
                             "some invariants")

        if isinstance(self, type):
            # If we are called as a classmethod, first get an instance
            # with all defaults set.  This will be our pattern.
            self = self(None, **kwargs)
            for invariant in invariants:
                value = self._header_parser.defaults[invariant]
                if value is None:
                    raise ValueError('can only set as invariant a header '
                                     'part that has a default.')
                self[invariant] = value

        # Create an all-zero version and set bits for all invariants.
        mask = self.__class__(None, **kwargs)
        for invariant in invariants:
            mask[invariant] = True

        return self.words, mask.words

    @fixedvalue
    def nbytes(cls):
        """Size of the header in bytes."""
        return cls._struct.size

    @classmethod
    def fromfile(cls, fh, *args, **kwargs):
        """Read VLBI Header from file.

        Arguments are the same as for class initialisation.  The header
        constructed will be immutable.
        """
        s = fh.read(cls._struct.size)
        if len(s) != cls._struct.size:
            raise EOFError
        return cls(cls._struct.unpack(s), *args, **kwargs)

    def tofile(self, fh):
        """Write VLBI frame header to filehandle."""
        return fh.write(self._struct.pack(*self.words))

    def _repr_value(self, key, value):
        if key.startswith(('bcd', 'crc', 'sync_pattern')):
            try:
                value = hex(value)
            except Exception:
                pass
        return super()._repr_value(key, value)



# Licensed under the GPLv3 - see LICENSE
"""
Definitions for VLBI VDIF frames and frame sets.

Implements a VDIFFrame class  that can be used to hold a header and a
payload, providing access to the values encoded in both.  Also, define
a VDIFFrameSet class that combines a set of frames from different threads.

For the VDIF specification, see https://www.vlbi.org/vdif
"""
import numpy as np

from ..base.frame import FrameBase
from .header import VDIFHeader, VDIFBaseHeader
from .payload import VDIFPayload


__all__ = ['VDIFFrame', 'VDIFFrameSet']


class VDIFFrame(FrameBase):
    """Representation of a VDIF data frame, consisting of a header and payload.

    Parameters
    ----------
    header : `~baseband.vdif.VDIFHeader`
        Wrapper around the encoded header words, providing access to the
        header information.
    payload : `~baseband.vdif.VDIFPayload`
        Wrapper around the payload, provding mechanisms to decode it.
    valid : bool, optional
        Whether the data are valid.  Default: inferred from header.
        Note that ``header`` is changed in-place if `True` or `False`.
    verify : bool
        Whether or not to do basic assertions that check the integrity
        (e.g., that channel information and whether or not data are complex
        are consistent between header and data).  Default: `True`.

    Notes
    -----
    The Frame can also be instantiated using class methods:

      fromfile : read header and payload from a filehandle

      fromdata : encode data as payload

    Of course, one can also do the opposite:

      tofile : method to write header and payload to filehandle

      data : property that yields full decoded payload

    One can decode part of the payload by indexing or slicing the frame.
    If the frame does not contain valid data, all values returned are set
    to ``self.fill_value``.

    A number of properties are defined: `shape`, `dtype` and `size` are
    the shape, type and number of complete samples of the data array, and
    `nbytes` the frame size in bytes.  Furthermore, the frame acts as a
    dictionary, with keys those of the header.  Any attribute that is not
    defined on the frame itself, such as ``.time`` will be looked up on the
    header as well.
    """

    _header_class = VDIFHeader
    _payload_class = VDIFPayload

    def verify(self):
        """Verify integrity.

        Checks consistency between the header information and payload
        data shape and type.
        """
        super().verify()
        assert self.header.complex_data == (self.payload.dtype.kind == 'c')
        assert self.payload.shape == (self.header.samples_per_frame,
                                      self.header.nchan)

    @property
    def valid(self):
        """Whether frame contains valid data.

        This is just the opposite of the ``invalid_data`` item in the header.
        If set, that header item is adjusted correspondingly.
        """
        return not self.header['invalid_data']

    @valid.setter
    def valid(self, valid):
        self.header['invalid_data'] = not valid

    @classmethod
    def fromfile(cls, fh, edv=None, verify=True):
        """Read a frame from a filehandle.

        Parameters
        ----------
        fh : filehandle
            From which the header and payload are read.
        edv : int, False, or None, optional
            Extended Data Version.  `False` is for legacy headers.  If `None`
            (default), it will be determined from the words themselves.
        verify : bool, optional
            Whether or not to do basic assertions that check the integrity
            (e.g., that channel information and whether or not data are complex
            are consistent between header and data).  Default: `True`.
        """
        header = cls._header_class.fromfile(fh, edv, verify)
        payload = cls._payload_class.fromfile(fh, header=header)
        # Since header was (optionally) verified, and payload was initialized
        # using header, frame verification is unnecessary.
        return cls(header, payload, verify=False)

    @classmethod
    def from_mark5b_frame(cls, mark5b_frame, verify=True, **kwargs):
        """Construct an Mark5B over VDIF frame (EDV=0xab).

        Any additional keywords can be used to set VDIF header properties
        not found in the Mark 5B header (such as station).

        See https://vlbi.org/wp-content/uploads/2019/03/vdif_extension_0xab.pdf
        """
        m5h, m5pl = mark5b_frame.header, mark5b_frame.payload
        header = cls._header_class.from_mark5b_header(
            m5h, nchan=m5pl.sample_shape.nchan, bps=m5pl.bps,
            invalid_data=not mark5b_frame.valid, **kwargs)
        payload = cls._payload_class(m5pl.words, header)
        return cls(header, payload, verify)


class VDIFFrameSet:
    """Representation of a set of VDIF frames, combining different threads.

    Parameters
    ----------
    frames : list of `~baseband.vdif.VDIFFrame`
        Should all cover the same time span.
    header0 : `~baseband.vdif.VDIFHeader`
        First header of the frame set.  If `None` (default), is extracted from
        ``frames[0]``.

    Notes
    -----
    The FrameSet can also be read instantiated using class methods:

      fromfile : read frames from a filehandle, optionally selecting threads

      fromdata : encode data as a set of frames

    Of course, one can also do the opposite:

      tofile : write frames to filehandle

      data : property that yields full decoded frame payloads

    One can decode part of the payload by indexing or slicing the frame.
    If the frame does not contain valid data, all values returned are set
    to ``self.fill_value``.

    A number of properties are defined: `shape`, `dtype` and `size` are
    the shape, type and number of complete samples of the data array, and
    `nbytes` the frame size in bytes.  Like a VDIFFrame, the frame set acts
    as a dictionary, with keys those of the header of the first frame
    (available via ``.header0``).  Any attribute that is not defined on the
    frame set itself, such as ``.time`` will also be looked up on the header.
    """

    def __init__(self, frames, header0=None):
        self.frames = frames
        if header0 is None:
            self.header0 = self.frames[0].header
        else:
            self.header0 = header0

    @classmethod
    def fromfile(cls, fh, thread_ids=None, edv=None, verify=True):
        """Read a frame set from a file, starting at the current location.

        Parameters
        ----------
        fh : filehandle
            Handle to the VDIF file.  Should be at the location where the
            frames are read from.
        thread_ids : list or None, optional
            The thread ids that should be read.  If `None` (default), continue
            reading threads as long as the frame number does not increase.
        edv : int or None, optional
            The expected extended data version for the VDIF Header.  If `None`
            (default), use that of the first frame.  (Passing it in slightly
            improves file integrity checking.)
        verify : bool, optional
            Whether to do (light) sanity checks on the header. Default: `True`.

        Returns
        -------
        frameset : `~baseband.vdif.VDIFFrameSet`
            Its ``frames`` property holds a list of frames (in order of either
            their ``thread_id`` or following the input ``thread_ids`` list).
            Use the `data` attribute to convert to an array.
        """
        header0 = VDIFHeader.fromfile(fh, edv, verify)
        edv = header0.edv
        frame_nr = header0['frame_nr']

        frames = {}
        header = header0
        while header['frame_nr'] == frame_nr:
            thread_id = header['thread_id']
            if thread_ids is None or thread_id in thread_ids:
                payload = VDIFPayload.fromfile(fh, header=header)
                # Since header was (optionally) verified, and payload was
                # initialized using header, no frame verification is needed.
                frames[thread_id] = VDIFFrame(header, payload, verify=False)
            else:
                fh.seek(header.payload_nbytes, 1)

            try:
                header = VDIFHeader.fromfile(fh, edv, verify)
            except (EOFError, AssertionError):
                if thread_ids is None or len(frames) == len(thread_ids):
                    break
                else:
                    raise

        else:  # Move back to before header that had incorrect frame_nr.
            fh.seek(-header.nbytes, 1)

        if thread_ids and len(frames) < len(thread_ids):
            raise OSError("could not find all requested frames.")

        # Turn dict of frames into a list, following order given by
        # thread_ids, or just sorting by their own thread_id
        if thread_ids is None:
            thread_ids = sorted(frames.keys())
        frames = [frames[tid] for tid in thread_ids]

        return cls(frames, header0)

    def tofile(self, fh):
        """Write all encoded frames to filehandle."""
        for frame in self.frames:
            frame.tofile(fh)

    @classmethod
    def fromdata(cls, data, headers=None, verify=True, **kwargs):
        """Construct a set of frames from data and headers.

        Parameters
        ----------
        data : `~numpy.ndarray`
            Array holding complex or real data to be encoded.  Dimensions
            should be (samples_per_frame, nthread, nchan).
        headers : `~baseband.vdif.VDIFHeader`, list of same, or None
            If a single header, a list with increasing ``thread_id`` is
            generated.  If not given, will attempt to generate a header from
            the keyword arguments.
        verify : bool
            Whether or not to do basic assertions that check the integrety
            (e.g., that channel information and whether or not data are complex
            are consistent between header and data).  Default: `True`.
        **kwargs
            If ``header`` is not given, these are used to initialize one.

        Returns
        -------
        frameset : `~baseband.vdif.VDIFFrameSet`
        """
        assert data.ndim == 3
        if not isinstance(headers, (list, tuple)):
            if headers is None:
                kwargs.setdefault('thread_id', 0)
                header = VDIFHeader.fromvalues(verify=verify, **kwargs)
            else:
                header = headers.copy()
            header['thread_id'] = 0
            headers = [header]
            for thread_id in range(1, data.shape[1]):
                header = header.copy()
                header['thread_id'] = thread_id
                headers.append(header)

        frames = [VDIFFrame.fromdata(d, h, verify=verify)
                  for d, h in zip(data.transpose(1, 0, 2), headers)]
        return cls(frames)

    @property
    def nbytes(self):
        """Size of the encoded frame in bytes."""
        return len(self.frames) * self.frames[0].nbytes

    @property
    def sample_shape(self):
        """Shape of a sample in the frameset (nthread, nchan)."""
        return (len(self.frames),) + self.frames[0].sample_shape

    def __len__(self):
        """Number of samples in the frameset."""
        return len(self.frames[0])

    @property
    def shape(self):
        """Shape of the frameset data."""
        return (len(self),) + self.sample_shape

    @property
    def size(self):
        """Total number of component samples in the frameset data."""
        prod = 1
        for dim in self.shape:
            prod *= dim
        return prod

    @property
    def ndim(self):
        """Number of dimensions of the frameset data."""
        return len(self.shape)

    @property
    def dtype(self):
        """Numeric type of the frameset data."""
        return self.frames[0].dtype

    @property
    def valid(self):
        """Whether frameset contains valid data."""
        valid = np.array([frame.valid for frame in self.frames])
        return valid[0] if len(np.unique(valid)) == 1 else valid

    @valid.setter
    def valid(self, valid):
        valid = np.broadcast_to(valid, (len(self.frames),))
        for f, v in zip(self.frames, valid):
            f.valid = v

    @property
    def fill_value(self):
        """Value to replace invalid data in the frameset."""
        return self.frames[0].fill_value

    @fill_value.setter
    def fill_value(self, fill_value):
        for frame in self.frames:
            frame.fill_value = fill_value

    def _get_frames(self, item):
        """Get frames and other information required to obtain given item.

        Parameters
        ----------
        item : int, slice, or tuple
            Sample indices.  An int represents a single sample, a slice a
            sample range, and a tuple of ints/slices a range for multi-frame
            and multi-channel data.

        Returns
        -------
        frames : list
            List of frames needed for this slice.  For a list of length unity,
            ``single_frame`` determines whether it is an index or unit-length
            slice.
        frame_item : int, slice, or tuple
            The item that should be gotten/set for each frame
        single_sample, single_frame, single_channel : bool
            Whether the sample, frame, or channel axes are simple indices,
            and thus whether the corresponding dimension should be removed.

        Notes
        -----
        The sample part of ``item`` is restricted to (tuples of) ints or
        slices, so one cannot access non-contiguous samples using advanced
        indexing.  Futhermore, if ``item`` is a slice, a negative increment
        cannot be used.  The function is unable to parse payloads whose words
        have unused space (eg. VDIF files with 20 bits/sample).
        """
        if item == ():
            return self.frames, (), False, False, False

        if not isinstance(item, tuple):
            return self.frames, item, not isinstance(item, slice), False, False

        single_sample = not isinstance(item[0], slice)
        if len(item) == 1:
            return self.frames, item[0], single_sample, False, False

        single_channel = (len(item) > 2
                          and np.empty(self.shape[2:])[item[2:]].ndim == 0)
        frame_indices = np.arange(len(self.frames))[item[1]]
        assert frame_indices.ndim <= 1
        single_frame = frame_indices.ndim == 0
        frames = [self.frames[i] for i in np.atleast_1d(frame_indices)]
        return (frames, item[:1] + item[2:],
                single_sample, single_frame, single_channel)

    # Header behaves as a dictionary, while Payload can be indexed/sliced.
    # Let frameset behave appropriately.
    def __getitem__(self, item=()):
        if isinstance(item, str):
            # Header behaves as a dictionary.  Assume base keywords are the
            # same for every frame, except for thread_id (must be different)
            # and invalid_data (can be different).
            if item == 'thread_id':
                return np.array([f.header[item] for f in self.frames])
            elif (item != 'invalid_data'
                  and item in VDIFBaseHeader._header_parser.keys()):
                return self.header0[item]

            values = np.array([f.header[item] for f in self.frames])
            return values[0] if len(np.unique(values)) == 1 else values

        (frames, frame_item,
         single_sample, single_frame, single_channel) = self._get_frames(item)

        data0 = frames[0][frame_item]
        if single_frame:
            return data0

        if single_sample:
            swapped = data = np.empty((len(frames),) + data0.shape,
                                      dtype=self.dtype)
        else:
            data = np.empty((data0.shape[0], len(frames))
                            + data0.shape[1:], dtype=self.dtype)
            swapped = data.swapaxes(0, 1)

        swapped[0] = data0
        for i, frame in enumerate(frames[1:]):
            swapped[i+1] = frame[frame_item]
        return data

    def __setitem__(self, item, data):
        if isinstance(item, str):
            # Headers behave as dictionaries; except for thread_id and
            # invalid_data, assume set base properties all to the same value.
            data = np.broadcast_to(data, (len(self.frames),))
            if item == 'thread_id':
                if len(np.unique(data)) != len(self.frames):
                    raise ValueError("all thread ids should be unique.")
            elif (item != 'invalid_data'
                  and item in VDIFBaseHeader._header_parser.keys()):
                if data.strides != (0,) and len(np.unique(data)) > 1:
                    raise ValueError("base header keys should be identical.")

            for f, value in zip(self.frames, data):
                f.header[item] = value
            return

        (frames, frame_item,
         single_sample, single_frame, single_channel) = self._get_frames(item)

        if single_frame:
            frames[0][frame_item] = data
            return

        data = np.asanyarray(data)
        if single_channel:
            if single_sample and data.ndim <= 1:
                swapped = np.broadcast_to(data, (len(frames),))
            else:
                new_shape = (data.shape[0], len(frames))
                swapped = np.broadcast_to(data, new_shape).swapaxes(0, 1)
        else:
            if single_sample or data.ndim <= 2:
                new_shape = (len(frames),) + data.shape[1:]
                swapped = np.broadcast_to(data, new_shape)
            else:
                new_shape = (data.shape[0], len(frames)) + data.shape[2:]
                swapped = np.broadcast_to(data, new_shape).swapaxes(0, 1)

        for frame, frame_data in zip(frames, swapped):
            frame[frame_item] = frame_data

    data = property(__getitem__, doc="Full decoded frame.")

    def keys(self):
        return self.header0.keys()

    def _ipython_key_completions_(self):
        # Enables tab-completion of header keys in IPython.
        return self.keys()

    def __contains__(self, key):
        return key in self.header0

    def __getattr__(self, attr):
        if attr in self.header0._properties:
            if attr in VDIFBaseHeader._properties:
                return getattr(self.header0, attr)
            values = np.hstack([getattr(f.header, attr) for f in self.frames])
            return values[0] if len(np.unique(values)) == 1 else values
        else:
            return self.__getattribute__(attr)

    # For tests, it is useful to define equality.
    def __eq__(self, other):
        return (type(self) is type(other)
                and len(self.frames) == len(other.frames)
                and self.header0 == other.header0
                and all(f1 == f2 for f1, f2 in zip(self.frames, other.frames)))



# Licensed under the GPLv3 - see LICENSE
"""Green Bank Ultimate Pulsar Processing Instrument (GUPPI) format
reader/writer.
"""
from .base import open, info, GUPPIFileNameSequencer  # noqa
from .header import GUPPIHeader  # noqa
from .payload import GUPPIPayload  # noqa
from .frame import GUPPIFrame  # noqa



# Licensed under the GPLv3 - see LICENSE
import importlib
import pytest
from astropy import units as u
from astropy.time import Time

from .. import file_info
from ..data import (SAMPLE_MARK4 as SAMPLE_M4, SAMPLE_MARK5B as SAMPLE_M5B,
                    SAMPLE_VDIF, SAMPLE_MWA_VDIF as SAMPLE_MWA, SAMPLE_DADA,
                    SAMPLE_PUPPI, SAMPLE_GSB_RAWDUMP_HEADER,
                    SAMPLE_GSB_RAWDUMP, SAMPLE_GSB_PHASED_HEADER,
                    SAMPLE_GSB_PHASED)


@pytest.mark.parametrize(
    ('sample', 'format_', 'missing', 'readable', 'error_keys'),
    ((SAMPLE_M4, 'mark4', True, True, []),
     (SAMPLE_M5B, 'mark5b', True, False, []),
     (SAMPLE_VDIF, 'vdif', False, True, []),
     (SAMPLE_MWA, 'vdif', False, True, ['frame_rate']),
     (SAMPLE_DADA, 'dada', False, True, []),
     (SAMPLE_PUPPI, 'guppi', False, True, []),
     (SAMPLE_GSB_RAWDUMP_HEADER, 'gsb', True, None, []),
     (SAMPLE_GSB_PHASED_HEADER, 'gsb', True, None, [])))
def test_basic_file_info(sample, format_, missing, readable, error_keys):
    info = file_info(sample)
    info_dict = info()
    assert info.format == format_
    assert info_dict['format'] == format_
    assert (hasattr(info, 'missing') and info.missing != {}) is missing
    assert ('missing' in info_dict and info_dict['missing'] != {}) is missing
    assert info.readable is readable
    assert list(info.errors.keys()) == error_keys


@pytest.mark.parametrize(
    ('sample', 'missing'),
    ((SAMPLE_M4, {'decade', 'ref_time'}),
     (SAMPLE_M5B, {'kday', 'ref_time', 'nchan'})))
def test_info_missing_args(sample, missing):
    info = file_info(sample)
    assert info.missing
    assert set(info.missing) == missing


@pytest.mark.parametrize(
    ('sample', 'format', 'wrong'),
    [(SAMPLE_M4, 'mark4', dict(decade='2010')),
     (SAMPLE_M5B, 'mark5b', dict(ref_time='56000', nchan=8))])
def test_info_wrong_args(sample, format, wrong):
    info = file_info(sample, **wrong)
    assert info.format == format
    assert not info.missing
    assert 'header0' in info.errors


@pytest.mark.parametrize(
    ('sample', 'format_', 'used', 'consistent', 'inconsistent'),
    ((SAMPLE_M4, 'mark4', ('ref_time',), ('nchan',), ()),
     (SAMPLE_M5B, 'mark5b', ('ref_time', 'nchan'), (), ()),
     (SAMPLE_VDIF, 'vdif', (), ('nchan', 'ref_time'), ()),
     (SAMPLE_DADA, 'dada', (), ('ref_time',), ('nchan',)),
     (SAMPLE_PUPPI, 'guppi', (), ('nchan',), ('ref_time',))))
def test_file_info(sample, format_, used, consistent, inconsistent):
    # Pass on extra arguments needed to get Mark4 and Mark5B to pass.
    # For GSB, we also need raw files, so we test that below.
    extra_args = {'ref_time': Time('2014-01-01'),
                  'nchan': 8}
    info = file_info(sample, **extra_args)
    assert info.format == format_
    info_dict = info()
    for attr in info.attr_names:
        info_value = getattr(info, attr)
        assert info_value is not None
        assert attr in info_dict or info_value == {}
    assert set(info.used_kwargs.keys()) == set(used)
    assert set(info.consistent_kwargs.keys()) == set(consistent)
    assert set(info.inconsistent_kwargs.keys()) == set(inconsistent)
    assert set(info.irrelevant_kwargs.keys()) == set()
    # Check that extraneous arguments get classified correctly.
    info2 = file_info(sample, life=42, **extra_args)
    assert info2.used_kwargs == info.used_kwargs
    assert info2.consistent_kwargs == info.consistent_kwargs
    assert info2.inconsistent_kwargs == info.inconsistent_kwargs
    assert info2.irrelevant_kwargs == {'life': 42}
    # Check we can indeed open a file with the extra arguments.
    module = importlib.import_module('.' + info.format, package='baseband')
    with module.open(sample, mode='rs', **info.used_kwargs) as fh:
        info3 = fh.info
    assert info3() == info_dict
    # Check that things properly do *not* work on a closed file.
    with module.open(sample, mode='rs', **info.used_kwargs) as fh:
        pass
    info4 = fh.info
    assert not info4
    assert 'File closed' in repr(info4)
    assert 'errors' in info4()
    assert any(isinstance(v, ValueError) for v in info4.errors.values())


@pytest.mark.parametrize(
    ('sample', 'raw', 'mode'),
    ((SAMPLE_GSB_RAWDUMP_HEADER, SAMPLE_GSB_RAWDUMP, 'rawdump'),
     (SAMPLE_GSB_PHASED_HEADER, SAMPLE_GSB_PHASED, 'phased')))
def test_gsb_with_raw_files(sample, raw, mode):
    # Note that the payload size in our sample files is reduced,
    # so without anything the file is not readable.
    bad_info = file_info(sample, raw=raw)
    assert bad_info.readable is False
    assert list(bad_info.errors.keys()) == ['frame0']
    # But with the correct sample_rate, it works.
    base_info = file_info(sample)
    sample_rate = (base_info.frame_rate
                   * (8192 if base_info.mode == 'rawdump' else 8))
    info = file_info(sample, raw=raw, sample_rate=sample_rate)
    assert info.format == 'gsb'
    assert info.readable is True
    assert not info.errors
    module = importlib.import_module('.' + info.format, package='baseband')
    # Check we can indeed open a file with the extra arguments.
    with module.open(sample, mode='rs', **info.used_kwargs) as fh:
        info2 = fh.info
    assert info2() == info()


def test_mwa_vdif_with_sample_rate():
    info1 = file_info(SAMPLE_MWA)
    assert info1.format == 'vdif'
    assert 'frame_rate' in info1.errors
    info2 = file_info(SAMPLE_MWA, sample_rate=1.28*u.MHz)
    assert info2.format == 'vdif'
    assert info2.start_time == Time('2015-10-03T20:49:45.000')
    info3 = file_info(SAMPLE_MWA, sample_rate='bla')
    assert info3.format == 'vdif'
    assert 'stream' in info3.errors


def test_unsupported_file(tmpdir):
    name = str(tmpdir.join('test.unsupported'))
    with open(name, 'wb') as fw:
        fw.write(b'abcdefghijklmnopqrstuvwxyz')

    info = file_info(name)
    assert not info
    assert 'does not seem formatted as any of' in str(info)

    info = file_info(name, format='vdif')
    assert 'errors' in str(info)
    assert 'Not parsable' in str(info)


def test_format_with_no_info(monkeypatch):
    # If a format does not provide info, we still try to open
    # with the stream opener, so the format should still get
    # recognized.  But we do note that the format checking was
    # haphazard.
    monkeypatch.delattr('baseband.vdif.info')
    info = file_info(SAMPLE_VDIF)
    assert info
    assert info.format == 'vdif'
    assert not hasattr(info, 'used_kwargs')

    info = file_info(SAMPLE_M4, format='vdif')
    assert not info
    assert "Error" in str(info)
    assert "no 'info'" in str(info)

    info = file_info(SAMPLE_M4, format=('vdif', 'mark5b'))
    assert not info
    assert "does not seem formatted as any of" in str(info)
    assert "({'vdif'} had no 'info'" in str(info)



"""
Definitions for VLBI Mark 4 Headers.

Implements a Mark4Header class used to store header words, and decode/encode
the information therein.

For the specification of tape Mark 4 format, see
https://www.haystack.mit.edu/tech/vlbi/mark5/docs/230.3.pdf

A little bit on the disk representation is at
https://ui.adsabs.harvard.edu/abs/2003ASPC..306..123W
"""
import struct

import numpy as np
from astropy.time import Time
from astropy.utils import sharedmethod

from ..base.header import HeaderParser, VLBIHeaderBase
from ..base.utils import bcd_decode, bcd_encode, fixedvalue, CRCStack

__all__ = ['CRC12', 'crc12', 'stream2words', 'words2stream',
           'Mark4TrackHeader', 'Mark4Header']


MARK4_DTYPES = {8: '<u1',  # this needs to start with '<' for words2stream.
                16: '<u2',
                32: '<u4',
                64: '<u8'}
"""Integer dtype used to encode a given number of tracks."""

PAYLOAD_NBITS = 20000
"""Number of bits per track per frame."""

CRC12 = 0x180f
"""CRC polynomial used for Mark 4 Headers.

x^12 + x^11 + x^3 + x^2 + x + 1, i.e., 0x180f.
See page 4 of https://www.haystack.mit.edu/tech/vlbi/mark5/docs/230.3.pdf

This is also a 'standard' CRC-12 mentioned in
https://en.wikipedia.org/wiki/Cyclic_redundancy_check
"""
crc12 = CRCStack(CRC12)


def stream2words(stream, track=None):
    """Convert a stream of integers to uint32 header words.

    Parameters
    ----------
    stream : `~numpy.array` of int
        For each int, every bit corresponds to a particular track.
    track : int, array, or None, optional
        The track to extract.  If `None` (default), extract all tracks that
        the type of int in the stream can hold.
    """
    if track is None:
        track = np.arange(stream.dtype.itemsize * 8, dtype=stream.dtype)

    track_sel = ((stream.reshape(-1, 32, 1) >> track) & 1).astype(np.uint32)
    track_sel <<= np.arange(31, -1, -1, dtype=np.uint32).reshape(-1, 1)
    words = np.bitwise_or.reduce(track_sel, axis=1)
    return words.squeeze()


def words2stream(words):
    """Convert a set of uint32 header words to a stream of integers.

    Parameters
    ----------
    words : `~numpy.array` of uint32

    Returns
    -------
    stream : `~numpy.array` of int
        For each int, every bit corresponds to a particular track.
    """
    ntrack = words.shape[1]
    dtype = MARK4_DTYPES[ntrack]
    nbits = words.dtype.itemsize * 8
    bit = np.arange(nbits - 1, -1, -1, dtype=words.dtype).reshape(-1, 1)

    bit_sel = ((words[:, np.newaxis, :] >> bit) & 1).astype(dtype[1:])
    bit_sel <<= np.arange(ntrack, dtype=dtype[1:])
    words = np.empty(bit_sel.shape[:2], dtype)
    words = np.bitwise_or.reduce(bit_sel, axis=2, out=words)
    return words.ravel()


class Mark4TrackHeader(VLBIHeaderBase):
    """Decoder/encoder of a Mark 4 Track Header.

    See https://www.haystack.mit.edu/tech/vlbi/mark5/docs/230.3.pdf

    Parameters
    ----------
    words : tuple of int, or None
        Five 32-bit unsigned int header words.  If `None`, set to a list of
        zeros for later initialisation.
    decade : int or None
        Decade in which the observations were taken (needed to remove ambiguity
        in the Mark 4 time stamp).  Can instead pass an approximate
        ``ref_time``.
    ref_time : `~astropy.time.Time` or None
        Reference time within 4 years of the observation time, used to infer
        the full Mark 4 timestamp.  Used only if ``decade`` is not given.
    verify : bool, optional
        Whether to do basic verification of integrity.  Default: `True`.

    Returns
    -------
    header : `~baseband.mark4.header.Mark4TrackHeader`
    """

    _header_parser = HeaderParser(
        (('bcd_headstack1', (0, 0, 16, 0x3344)),
         ('bcd_headstack2', (0, 16, 16, 0x1122)),
         ('headstack_id', (1, 30, 2)),
         ('bcd_track_id', (1, 24, 6)),
         ('fan_out', (1, 22, 2)),
         ('magnitude_bit', (1, 21, 1)),
         ('lsb_output', (1, 20, 1)),
         ('converter_id', (1, 16, 4)),
         ('time_sync_error', (1, 15, 1, False)),
         ('internal_clock_error', (1, 14, 1, False)),
         ('processor_time_out_error', (1, 13, 1, False)),
         ('communication_error', (1, 12, 1, False)),
         ('_1_11_1', (1, 11, 1, False)),
         ('_1_10_1', (1, 10, 1, False)),
         ('track_roll_enabled', (1, 9, 1, False)),
         ('sequence_suspended', (1, 8, 1, False)),
         ('system_id', (1, 0, 8)),
         ('_1_0_1_sync', (1, 0, 1, 0)),  # Lowest bit of system ID is 0.
         ('sync_pattern', (2, 0, 32, 0xffffffff)),
         ('bcd_unit_year', (3, 28, 4)),
         ('bcd_day', (3, 16, 12)),
         ('bcd_hour', (3, 8, 8)),
         ('bcd_minute', (3, 0, 8)),
         ('bcd_second', (4, 24, 8)),
         ('bcd_fraction', (4, 12, 12)),
         ('crc', (4, 0, 12))))
    _sync_pattern = _header_parser.defaults['sync_pattern']
    _invariants = {'sync_pattern', '_1_0_1_sync'}
    """Keys of invariant parts in all Mark 4 headers.

    This includes the lowest bit of 'system_id', which is apparently
    always 0 (at least, mark5access assumes so too).
    """
    _stream_invariants = (_invariants
                          | {'bcd_headstack1', 'bcd_headstack2',
                             'track_roll_enabled', 'sequence_suspended',
                             'system_id'})
    """Keys of invariant parts in a given Mark 4 stream."""

    _struct = struct.Struct('<5I')

    _properties = ('decade', 'track_id', 'fraction', 'time')
    """Properties accessible/usable in initialisation."""

    decade = None
    """Decade of year, to complement 'bcd_unit_year' from header."""

    def __init__(self, words, decade=None, ref_time=None, verify=True):
        if decade is not None:
            self.decade = decade
        super().__init__(words, verify=verify)
        if decade is None and ref_time is not None:
            self.infer_decade(ref_time)

    def verify(self):
        """Verify header integrity."""
        assert len(self.words) == 5
        assert np.all(self['sync_pattern'] == self._sync_pattern)
        assert np.all((self['bcd_fraction'] & 0xf) % 5 != 4)
        if self.decade is not None:
            assert (1950 < self.decade < 3000)
            assert self.decade % 10 == 0, "decade must end in zero"

    def infer_decade(self, ref_time):
        """Uses a reference time to set a header's ``decade``.

        Parameters
        ----------
        ref_time : `~astropy.time.Time`
            Reference time within 5 years of the observation time.
        """
        self.decade = np.around(ref_time.decimalyear - self['bcd_unit_year'],
                                decimals=-1).astype(int)

    @property
    def track_id(self):
        """Track identifier (decoded from 'bcd_track_id')."""
        return bcd_decode(self['bcd_track_id'])

    @track_id.setter
    def track_id(self, track_id):
        self['bcd_track_id'] = bcd_encode(track_id)

    @property
    def fraction(self):
        """Fractional seconds (decoded from 'bcd_fraction')."""
        ms = bcd_decode(self['bcd_fraction'])
        # The last digit encodes a fraction -- see table 2 in
        # https://www.haystack.mit.edu/tech/vlbi/mark5/docs/230.3.pdf
        # 0: 0.00      5: 5.00
        # 1: 1.25      6: 6.25
        # 2: 2.50      7: 7.50
        # 3: 3.75      8: 8.75
        # 4: invalid   9: invalid
        last_digit = ms % 5
        return (ms + last_digit * 0.25) / 1000.

    @fraction.setter
    def fraction(self, fraction):
        ms = fraction * 1000.
        if np.any(np.abs((ms / 1.25) - np.around(ms / 1.25)) > 1e-6):
            raise ValueError("{0} ms is not a multiple of 1.25 ms"
                             .format(ms))
        self['bcd_fraction'] = bcd_encode(np.floor(ms + 1e-6)
                                          .astype(np.int32))

    def get_time(self):
        """Convert BCD time code to Time object.

        Calculate time using bcd-encoded 'bcd_unit_year', 'bcd_day',
        'bcd_hour', 'bcd_minute', 'bcd_second' header items, as well as
        the ``fraction`` property (inferred from 'bcd_fraction') and
        ``decade`` from the initialisation.  See
        See https://www.haystack.mit.edu/tech/vlbi/mark5/docs/230.3.pdf
        """
        return Time('{decade:03d}{uy:1x}:{d:03x}:{h:02x}:{m:02x}:{s:08.5f}'
                    .format(decade=self.decade//10, uy=self['bcd_unit_year'],
                            d=self['bcd_day'], h=self['bcd_hour'],
                            m=self['bcd_minute'],
                            s=bcd_decode(self['bcd_second']) + self.fraction),
                    format='yday', scale='utc', precision=5)

    def set_time(self, time):
        """Convert Time object to BCD timestamp elements.

        Parameters
        ----------
        time : `~astropy.time.Time`
            The time to use for this header.
        """
        old_precision = time.precision
        try:
            time.precision = 5
            yday = time.yday.split(':')
        finally:
            time.precision = old_precision
        # Set fraction first since that checks precision.
        self.fraction = float(yday[4]) % 1
        self.decade = int(yday[0][:3]) * 10
        self['bcd_unit_year'] = int(yday[0][3], base=16)
        self['bcd_day'] = int(yday[1], base=16)
        self['bcd_hour'] = int(yday[2], base=16)
        self['bcd_minute'] = int(yday[3], base=16)
        self['bcd_second'] = int(yday[4][:2], base=16)

    time = property(get_time, set_time)


class Mark4Header(Mark4TrackHeader):
    """Decoder/encoder of a Mark 4 Header, containing all streams.

    See https://www.haystack.mit.edu/tech/vlbi/mark5/docs/230.3.pdf

    Parameters
    ----------
    words : `~numpy.ndarray` of int, or None
        Shape should be (5, number-of-tracks), and dtype np.uint32.  If `None`,
        ``ntrack`` should be given and words will be initialized to 0.
    ntrack : None or int
        Number of Mark 4 bitstreams, to help initialize ``words`` if needed.
    decade : int or None
        Decade in which the observations were taken (needed to remove ambiguity
        in the Mark 4 time stamp).  Can instead pass an approximate
        ``ref_time``.
    ref_time : `~astropy.time.Time` or None
        Reference time within 4 years of the observation time, used to infer
        the full Mark 4 timestamp.  Used only if ``decade`` is not given.
    verify : bool, optional
        Whether to do basic verification of integrity.  Default: `True`.

    Returns
    -------
    header : `~baseband.mark4.Mark4Header`
    """

    _track_header = Mark4TrackHeader
    _properties = (Mark4TrackHeader._properties
                   + ('fanout', 'samples_per_frame', 'bps', 'complex_data',
                      'nchan', 'sample_shape', 'nsb', 'converters'))
    _dtypes = MARK4_DTYPES

    # keyed with bps, fanout; Tables 10-14 in reference documentation:
    # https://www.haystack.mit.edu/tech/vlbi/mark5/docs/230.3.pdf
    # rows are channels with Sign, Mag for each for bps=2, columns fanout.
    # So for bps=2, fanout=4 (abbreviating channel a Sign, Mag as aS, aM):
    # Channel a has samples (aS, aM) in tracks (2, 10), (4,12), etc.
    #         b             (bS, bM) in        (3, 11), etc.
    # We subtract two and reshape as (fanout, nchan, bps) since that is how
    # it is used internally.
    _track_assignments = {
        (2, 4): np.array(  # rows=aS, aM, bS, bM, cS, cM, dS, dM; cols=fanout.
            [[2, 10, 3, 11, 18, 26, 19, 27],
             [4, 12, 5, 13, 20, 28, 21, 29],
             [6, 14, 7, 15, 22, 30, 23, 31],
             [8, 16, 9, 17, 24, 32, 25, 33]]).reshape(4, 4, 2) - 2,
        (1, 4): np.array(  # rows=aS, bS, ..., hS; cols=fanout.
            [[2, 3, 10, 11, 18, 19, 26, 27],
             [4, 5, 12, 13, 20, 21, 28, 29],
             [6, 7, 14, 15, 22, 23, 30, 31],
             [8, 9, 16, 17, 24, 25, 32, 33]]).reshape(4, 8, 1) - 2,
        (2, 2): (np.array(  # rows=aS, aM, bS, bM, ..., hS, hM; cols=fanout.
            [[2, 6, 3, 7, 10, 14, 11, 15, 18, 22, 19, 23, 26, 30, 27, 31],
             [4, 8, 5, 9, 12, 16, 13, 17, 20, 24, 21, 25, 28, 32, 29, 33]])
                 .reshape(2, 8, 2) - 2),
        (1, 2): (np.array(  # rows=aS, bS, ..., pS; cols=fanout.
            [[2, 3, 6, 7, 10, 11, 14, 15, 18, 19, 22, 23, 26, 27, 30, 31],
             [4, 5, 8, 9, 12, 13, 16, 17, 20, 21, 24, 25, 28, 29, 32, 33]])
                 .reshape(2, 16, 1) - 2),
        (2, 1): (np.array(  # rows=aS, aM, bS, bM, ..., pS, pM; no fanout.
            [[2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22, 24, 26, 28, 30, 32,
              3, 5, 7, 9, 11, 13, 15, 17, 19, 21, 23, 25, 27, 29, 31, 33]])
                 .reshape(1, 16, 2) - 2)}

    def __init__(self, words, ntrack=None, decade=None, ref_time=None,
                 verify=True):
        if words is None:
            words = np.zeros((5, ntrack), dtype=np.uint32)
            verify = False
        super().__init__(words, decade=decade, ref_time=ref_time,
                         verify=verify)

    def verify(self):
        super().verify()
        assert set(self['fan_out']) == set(np.arange(self.fanout))
        assert (len(set((c, l) for (c, l) in zip(self['converter_id'],
                                                 self['lsb_output'])))
                == self.nchan)

    @sharedmethod
    def invariant_pattern(self, invariants=None, ntrack=None):
        """Invariant pattern to help search for headers.

        On the class, like mark5access, we use use one bit more than the sync
        pattern in word 2, viz., lsb of word 1, which we assume is always 0
        (it is the lowest bit of eight of 'system_id').

        Parameters
        ----------
        invariants : set of str, optional
            Set of keys to header parts that are shared between all headers
            of a given type or within a given stream/file.  Default: from
            `~baseband.base.header.VLBIHeaderBase.invariants()`.
        ntrack : int, optional
            Number of tracks.  Required for getting class invariants,
            ignored for instances.
        """

        if not isinstance(self, type):
            ntrack = self.ntrack

        elif ntrack is None:
            raise ValueError("need to pass in ``ntrack`` to "
                             "get Mark 4 generic invariants.")

        pattern, mask = super().invariant_pattern(invariants=invariants,
                                                  ntrack=ntrack)
        return words2stream(pattern), words2stream(mask)

    def infer_decade(self, ref_time):
        super().infer_decade(ref_time)
        if getattr(self.decade, 'size', 1) > 1:
            assert np.all(self.decade == self.decade[0])
            self.decade = self.decade[0]

    @classmethod
    def _stream_dtype(cls, ntrack):
        return np.dtype(cls._dtypes[ntrack])

    @property
    def stream_dtype(self):
        """Stream dtype required to hold this header's number of tracks."""
        return self._stream_dtype(self.ntrack)

    @classmethod
    def _track_assignment(cls, ntrack, bps, fanout):
        try:
            ta = cls._track_assignments[(bps, fanout)]
        except KeyError:
            raise ValueError("Mark 4 reader does not support bps={0}, "
                             "fanout={1}; supported are {2}".format(
                                 bps, fanout, cls._track_assignments.keys()))

        if ntrack == 64:
            # double up the number of tracks and channels.
            return np.concatenate((ta, ta + 32), axis=1)
        elif ntrack == 32:
            return ta
        elif ntrack == 16:
            return ta[:, ::2, :] // 2
        else:
            raise ValueError("have Mark 4 track assignments only for "
                             "ntrack=32 or 64, not {0}".format(ntrack))

    @property
    def track_assignment(self):
        """Assignments of tracks to channels and fanout items.

        The assignments are inferred from tables 10-14 in
        https://www.haystack.mit.edu/tech/vlbi/mark5/docs/230.3.pdf
        except that 2 has been subtracted so that tracks start at 0,
        and that for 64 tracks the arrays are suitably enlarged by adding
        another set of channels.

        The returned array has shape ``(fanout, nchan, bps)``.
        """
        return self._track_assignment(self.ntrack, self.bps, self.fanout)

    @classmethod
    def fromfile(cls, fh, ntrack, decade=None, ref_time=None, verify=True):
        """Read Mark 4 header from file.

        Parameters
        ----------
        fh : filehandle
            To read header from.
        ntrack : int
            Number of Mark 4 bitstreams.
        decade : int or None
            Decade in which the observations were taken.  Can instead pass an
            approximate ``ref_time``.
        ref_time : `~astropy.time.Time` or None
            Reference time within 4 years of the observation time.  Used only
            if ``decade`` is not given.
        verify : bool, optional
            Whether to do basic verification of integrity.  Default: `True`.
        """
        dtype = cls._stream_dtype(ntrack)
        header_nbytes = ntrack * 160 // 8
        try:
            stream = np.frombuffer(fh.read(header_nbytes), dtype=dtype)
            assert len(stream) * dtype.itemsize == header_nbytes
        except (ValueError, AssertionError):
            raise EOFError("could not read full Mark 4 Header.")

        words = stream2words(stream)
        self = cls(words, decade=decade, ref_time=ref_time, verify=verify)
        self.mutable = False
        return self

    def tofile(self, fh):
        stream = words2stream(self.words)
        fh.write(stream.tobytes())

    @classmethod
    def fromvalues(cls, ntrack, decade=None, ref_time=None, **kwargs):
        """Initialise a header from parsed values.

        Here, the parsed values must be given as keyword arguments, i.e., for
        any ``header = cls(<words>)``, ``cls.fromvalues(**header) == header``.

        However, unlike for the `fromkeys` class method, data can also be set
        using arguments named after header methods, such as ``time``.

        Parameters
        ----------
        ntrack : int
            Number of Mark 4 bitstreams.
        decade : int or None, optional
            Decade in which the observations were taken.  Can instead pass an
            approximate ``ref_time``.  Not needed if ``time`` is given.
        ref_time : `~astropy.time.Time` or None, optional
            Reference time within 4 years of the observation time.  Used only
            if ``decade`` is not given, and not needed if ``time`` is given.
        **kwargs :
            Values used to initialize header keys or methods.

        --- Header keywords : (minimum for a complete header)

        time : `~astropy.time.Time` instance
            Time of the first sample.
        bps : int
            Bits per elementary sample.
        fanout : int
            Number of tracks over which a given channel is spread out.
        """
        # set defaults based on ntrack for cases where it is known.
        if ntrack == 64:
            kwargs.setdefault('headstack_id', np.repeat(np.arange(2), 32))
            kwargs.setdefault('track_id', np.tile(np.arange(2, 34), 2))
        elif ntrack == 32:
            kwargs.setdefault('headstack_id', np.zeros(32, dtype=int))
            kwargs.setdefault('track_id', np.arange(2, 34))
        elif ntrack == 16:
            kwargs.setdefault('headstack_id', np.zeros(16, dtype=int))
            kwargs.setdefault('track_id', np.arange(2, 34, 2))
        # set number of sidebands to default if no information is given,
        # so that the header will be valid.
        if not any(key in kwargs for key in ('lsb_output', 'converter_id',
                                             'converter')):
            kwargs.setdefault('nsb', 1)
        return super().fromvalues(ntrack, decade, ref_time, **kwargs)

    def update(self, crc=None, verify=True, **kwargs):
        """Update the header by setting keywords or properties.

        Here, any keywords matching header keys are applied first, and any
        remaining ones are used to set header properties, in the order set
        by the class (in ``_properties``).

        Parameters
        ----------
        crc : int or None, optional
            If `None` (default), recalculate the CRC after updating.
        verify : bool, optional
            If `True` (default), verify integrity after updating.
        **kwargs
            Arguments used to set keywords and properties.
        """
        if crc is None:
            super().update(verify=False, **kwargs)
            stream = words2stream(self.words)
            stream[-12:] = crc12(stream[:-12])
            self.words = stream2words(stream)
            if verify:
                self.verify()
        else:
            super().update(verify=verify, crc=crc, **kwargs)

    @property
    def ntrack(self):
        """Number of Mark 4 bitstreams."""
        return self.words.shape[1]

    @property
    def nbytes(self):
        """Size of the header in bytes."""
        return self.ntrack * 160 // 8

    @property
    def frame_nbytes(self):
        """Size of the frame in bytes."""
        return self.ntrack * PAYLOAD_NBITS // 8

    @property
    def payload_nbytes(self):
        """Size of the payload in bytes.

        Note that the payloads miss pieces overwritten by the header.
        """
        return self.frame_nbytes - self.nbytes

    @property
    def fanout(self):
        """Number of samples stored in one payload item of size ntrack.

        If set, will update 'fan_out' for each track.
        """
        return np.max(self['fan_out']) + 1

    @fanout.setter
    def fanout(self, fanout):
        if fanout not in (1, 2, 4):
            raise ValueError("Mark 4 data only supports fanout=1, 2, or 4, "
                             "not {0}.".format(fanout))
        # In principle, one would like to go through track_assignments, but
        # we may not have bps set here yet, so just infer from tables:
        # fanout = 4: (0,1,2,3) * ntrack / 4              if ntrack = 16
        #             (0,0,1,1,2,2,3,3) * ntrack / 2 / 4  otherwise
        # fanout = 2: (0,0,1,1) * ntrack / 2 / 2
        # fanout = 1: (0,0) * ntrack / 2
        if self.ntrack == 16:
            self['fan_out'] = np.tile(np.arange(fanout), self.ntrack // fanout)
        else:
            self['fan_out'] = np.tile(np.repeat(np.arange(fanout), 2),
                                      self.ntrack // 2 // fanout)

    @property
    def samples_per_frame(self):
        """Number of complete samples in the frame.

        If set, this uses the number of tracks to infer and set `fanout`.
        """
        # Header overwrites part of payload, so we need
        # frame_nbytes * 8 // bps // nchan, but use ntrack and fanout, as these
        # are more basic; ntrack / fanout by definition equals bps * nchan.
        return self.frame_nbytes * 8 // (self.ntrack // self.fanout)

    @samples_per_frame.setter
    def samples_per_frame(self, samples_per_frame):
        fanout, extra = divmod(samples_per_frame * self.ntrack,
                               8 * self.frame_nbytes)
        if extra or fanout not in (1, 2, 4):
            raise ValueError(
                "header cannot store {} samples per frame. "
                "Should be one of {}."
                .format(samples_per_frame,
                        ', '.join([str(f * 8 * self.frame_nbytes)
                                   for f in (1, 2, 4)])))
        self.fanout = int(fanout)

    @property
    def bps(self):
        """Bits per elementary sample (either 1 or 2).

        If set, combined with `fanout` and `ntrack` to update 'magnitude_bit'
        for all tracks.
        """
        return 2 if self['magnitude_bit'].any() else 1

    @bps.setter
    def bps(self, bps):
        if bps == 1:
            self['magnitude_bit'] = False
        elif bps == 2:
            # Note: cannot assign to slice of header property, so go via array.
            ta = self._track_assignment(self.ntrack, bps, self.fanout)
            magnitude_bit = np.empty(self.ntrack, dtype=bool)
            magnitude_bit[ta] = [False, True]
            self['magnitude_bit'] = magnitude_bit
        else:
            raise ValueError("Mark 4 data can only have bps=1 or 2, "
                             "not {0}".format(bps))

    @fixedvalue
    def complex_data(cls):
        """Whether the data are complex.  Always False for Mark 4."""
        return False

    @property
    def nchan(self):
        """Number of channels (``ntrack * fanout``) in the frame.

        If set, it is combined with `ntrack` and `fanout` to infer `bps`.
        """
        return self.ntrack // (self.fanout * self.bps)

    @nchan.setter
    def nchan(self, nchan):
        self.bps = self.ntrack // (self.fanout * nchan)

    @property
    def sample_shape(self):
        """Shape of a sample in the payload (nchan,)."""
        return (self.nchan,)

    @sample_shape.setter
    def sample_shape(self, sample_shape):
        self.nchan, = sample_shape

    @property
    def nsb(self):
        """Number of side bands used.

        If set, assumes all converters are upper sideband for 1, and that
        converter IDs alternate between upper and lower sideband for 2.
        """
        sb = self['lsb_output']
        return 1 if (sb == sb[0]).all() else 2

    @nsb.setter
    def nsb(self, nsb):
        if nsb == 1:
            self['lsb_output'] = True

        elif nsb == 2:
            ta = self.track_assignment
            ta_ch = ta[0, :, 0]
            sb = np.tile([False, True], len(ta_ch) // 2)
            lsb_output = np.empty(self.ntrack, bool)
            lsb_output[ta] = sb[:, np.newaxis]
            self['lsb_output'] = np.tile([False, True], 16)

        else:
            raise ValueError("number of sidebands can only be 1 or 2.")

        # Set default converters; can be overridden if needed.
        nconverter = self.ntrack // (self.fanout * self.bps * self.nsb)
        converters = np.arange(nconverter)
        if nconverter > 2:
            converters = (converters.reshape(-1, 2, 2)
                          .transpose(0, 2, 1).ravel())
        self.converters = converters

    @property
    def converters(self):
        """Converted ID and sideband used for each channel.

        Returns a structured array with numerical 'converter' and boolean
        'lsb' entries (where `True` means lower sideband).

        Can be set with a similar structured array or a `dict`; if just an
        an array is passed in, it will be assumed that the sideband has been
        set beforehand (e.g., by setting `nsb`) and that the array holds
        the converter IDs.
        """
        ta_ch = self.track_assignment[0, :, 0]
        converters = np.empty(len(ta_ch), [("converter", int), ("lsb", bool)])
        converters['converter'] = self['converter_id'][ta_ch]
        converters['lsb'] = self['lsb_output'][ta_ch]
        return converters

    @converters.setter
    def converters(self, converters):
        # Set converters, duplicating over fanout, lsb, magnitude bit.
        ta = self.track_assignment
        ta_ch = ta[0, :, 0]
        nchan = len(ta_ch)
        msg = ('Mark 4 file with bps={0}, fanout={1} '
               'needs to define {2} converters')
        try:
            converter = converters['converter']
        except(KeyError, ValueError, IndexError):
            converter = np.array(converters)
            sb = self['lsb_output'][ta_ch]
            if self.nsb == 2 and len(converter) == len(ta_ch) // 2:
                c = np.empty(len(ta_ch), dtype=int)
                c[sb] = c[~sb] = converter
                converter = c
            if len(converter) != nchan:
                raise ValueError(msg.format(self.bps, self.fanout, nchan))

        else:
            sb = np.array(converters['lsb'])
            if len(converter) != nchan:
                raise ValueError(msg.format(self.bps, self.fanout, nchan))
            lsb_output = np.empty(self.ntrack, bool)
            lsb_output[ta] = sb[:, np.newaxis]
            self['lsb_output'] = lsb_output

        # Note: cannot assign to slice of header property, so go via array.
        converter_id = np.empty(self.ntrack, dtype=int)
        converter_id[ta] = converter[:, np.newaxis]
        self['converter_id'] = converter_id

    def get_time(self):
        """Convert BCD time code to Time object for all tracks.

        If all tracks have the same fractional seconds, only a single Time
        instance is returned.

        Uses bcd-encoded 'unit_year', 'day', 'hour', 'minute', 'second' and
        'frac_sec', plus ``decade`` from the initialisation to calculate the
        time.  See https://www.haystack.mit.edu/tech/vlbi/mark5/docs/230.3.pdf
        """
        if len(set(self['bcd_fraction'])) == 1:
            return self[0].time
        else:
            return Time([h.time for h in self], precision=5)

    def set_time(self, time):
        if time.isscalar:
            super().set_time(time)
        else:
            decades = set()
            for h, t in zip(self, time):
                h.set_time(t)
                decades.add(h.decade)
            if len(decades) > 1:
                raise ValueError("Mark4Header cannot have tracks that differ "
                                 "in the decade of the time they were taken.")
            self.decade = decades.pop()

    time = property(get_time, set_time)

    def __len__(self):
        return self.ntrack

    def __getitem__(self, item):
        if isinstance(item, str):
            return super().__getitem__(item)

        try:
            new_words = self.words[:, item]
        except IndexError:
            raise IndexError("index {item} is out of bounds.")

        if not(1 <= new_words.ndim <= 2 and new_words.shape[0] == 5):
            raise ValueError("cannot extract {0} from {1} instance."
                             .format(item, type(self)))

        if new_words.ndim == 1:
            return self._track_header(new_words, self.decade,
                                      verify=False)
        else:
            return self.__class__(new_words, self.decade, verify=False)

    def __eq__(self, other):
        return (type(self) is type(other)
                and np.all(self.words == other.words))

    def __repr__(self):
        name = self.__class__.__name__
        outs = []
        for k in self.keys():
            v = self[k]
            if len(v) == 1:
                repr_value = self._repr_value(k, v[0])
            elif np.all(v == v[0]):
                repr_value = f'[{self._repr_value(k, v[0])}]*{v.size}'
            else:
                if len(v) > 4:
                    v = (v[0], '...', v[-1])
                repr_value = '[{}]'.format(', '.join([self._repr_value(k, _v)
                                                      for _v in v]))
            outs.append(f'{k}: {repr_value}')

        return "<{} {}>".format(name, (",\n  " + " "*len(name)).join(outs))



# Licensed under the GPLv3 - see LICENSE
"""
Definitions for GUPPI headers.

Implements a GUPPIHeader class that reads & writes FITS-like headers from file.
"""
import operator

import numpy as np
import astropy.units as u
from astropy.io import fits
from astropy.time import Time


__all__ = ['GUPPIHeader']


class GUPPIHeader(fits.Header):
    """GUPPI baseband file format header.

    Parameters
    ----------
    *args : str or iterable
        If a string, parsed as a GUPPI header from a file, otherwise
        as for the `astropy.io.fits.Header` baseclass.
    verify : bool, optional
        Whether to do minimal verification that the header is consistent with
        the GUPPI standard.  Default: `True`.
    mutable : bool, optional
        Whether to allow the header to be changed after initialisation.
        Default: `True`.
    **kwargs
        Any further header keywords to be set.

    Notes
    -----
    Like `~astropy.io.fits.Header`, the initialiser does not accept keyword
    arguments to populate an array - instead, one must pass an iterable. In
    order to ensure keywords are kept in the right order, one should pass on
    values as a tuple, not as a dict. E.g., to copy a header, one should not
    do ``GUPPIHeader({key: header[key] for key in header})``, but rather::

        GUPPIHeader(((key, header[key]) for key in header))

    or, to also keep the comments::

        GUPPIHeader(((key, (header[key], header.comments[key]))
                      for key in header))
    """

    _properties = ('payload_nbytes', 'frame_nbytes', 'bps', 'nchan', 'npol',
                   'sample_shape', 'sample_rate', 'sideband', 'overlap',
                   'samples_per_frame', 'offset', 'start_time',
                   'time')
    """Properties accessible/usable in initialisation for all headers."""

    _defaults = [('BACKEND', 'GUPPI'),
                 ('BLOCSIZE', 0),
                 ('STT_OFFS', 0),
                 ('PKTIDX', 0),
                 ('OVERLAP', 0),
                 ('SRC_NAME', 'unset'),
                 ('TELESCOP', 'unset'),
                 ('PKTFMT', '1SFA'),
                 ('PKTSIZE', 8192),
                 ('NBITS', 8),
                 ('NPOL', 1),
                 ('OBSNCHAN', 1)]

    def __init__(self, *args, verify=True, mutable=True, **kwargs):
        # Comments handled by fits.Header__init__().
        super().__init__(*args, **kwargs)
        self.mutable = mutable
        # Empty header always OK, since things will be added to it.
        if len(self) and verify:
            self.verify()

    def verify(self):
        """Basic check of integrity."""
        assert all(key in self for key in ('BLOCSIZE',
                                           'PKTIDX'))
        # '1SFA' is used for all modes other than FAST4K (which is only used
        # for total intensity).  'SIMPLE' is from DSPSR, and used to support
        # time-first payloads.  See
        # https://safe.nrao.edu/wiki/pub/Main/JoeBrandt/guppi_status_shmem.pdf
        assert self['PKTFMT'] in ('1SFA', 'SIMPLE')

    def copy(self):
        """Create a mutable and independent copy of the header."""
        # This method exists because io.fits.Header.copy has a docstring that
        # refers to `Header` which breaks sphinx...
        return super().copy()

    @classmethod
    def fromfile(cls, fh, verify=True):
        """Reads in GUPPI header block from a file.

        Parameters
        ----------
        fh : filehandle
            To read data from.
        verify: bool, optional
            Whether to do basic checks on whether the header is valid. Verify
            is automatically called by `~astropy.io.fits.Header.fromstring`, so
            this flag exists only to standardize the API.
        """
        header_start = fh.tell()
        # Find the size of the header.  GUPPI header entries are 80 char long
        # with <=8 char keyword names.  "=" is always the 9th char.
        line = '========='
        while line[8] in ('=', ' '):
            line = fh.read(80).decode('ascii')
            if line[:3] == 'END':
                break
            if line == '':
                raise EOFError

        header_end = fh.tell()
        fh.seek(header_start)
        hdr = fh.read(header_end - header_start).decode('ascii')
        # Create the header using the base class.
        self = cls.fromstring(hdr)
        self.mutable = False
        if verify:
            self.verify()
        # GUPPI headers are not a proper FITS standard, and we're reading
        # from a file that the user likely cannot control, so let's not bother
        # with card verification (this avoids warnings in repr(); gh-282)
        for c in self.cards:
            c._verified = True
        return self

    def tofile(self, fh):
        """Write GUPPI file header to filehandle.

        Uses `~astropy.io.fits.Header.tostring`.
        """
        fh.write(self.tostring(padding=False).encode('ascii'))

    @classmethod
    def fromkeys(cls, *args, verify=True, mutable=True, **kwargs):
        """Initialise a header from keyword values.

        Like fromvalues, but without any interpretation of keywords.

        Note that this just passes kwargs to the class initializer as a dict
        (for compatibility with fits.Header). It is present for compatibility
        with other header classes only.
        """
        return cls(kwargs, *args, verify=verify, mutable=mutable)

    @classmethod
    def fromvalues(cls, **kwargs):
        """Initialise a header from parsed values.

        Here, the parsed values must be given as keyword arguments, i.e., for
        any ``header``, ``cls.fromvalues(**header) == header``.

        However, unlike for the ``fromkeys`` class method, data can also be set
        using arguments named after header methods, such as ``time``.

        Furthermore, some header defaults are set in ``GUPPIHeader._defaults``.
        """
        self = cls(cls._defaults, verify=False)
        self.update(**kwargs)
        return self

    def update(self, *, verify=True, **kwargs):
        """Update the header with new values.

        Here, any keywords matching properties are processed as well, in the
        order set by the class (in ``_properties``), and after all other
        keywords have been processed.

        Parameters
        ----------
        verify : bool, optional
            If `True` (default), verify integrity after updating.
        **kwargs
            Arguments used to set keywords and properties.
        """
        # Remove kwargs that set properties, in correct order.
        extras = [(key, kwargs.pop(key)) for key in self._properties
                  if key in kwargs]
        # Update the normal keywords.
        super().update(kwargs)
        # Now set the properties.
        for attr, value in extras:
            setattr(self, attr, value)
        if verify:
            self.verify()

    def __setitem__(self, key, value):
        if not self.mutable:
            raise TypeError("immutable {0} does not support assignment."
                            .format(type(self).__name__))

        super().__setitem__(key.upper(), value)

    @property
    def nbytes(self):
        """Size of the header in bytes."""
        return (len(self) + 1) * 80

    @property
    def payload_nbytes(self):
        """Size of the payload in bytes."""
        return self['BLOCSIZE']

    @payload_nbytes.setter
    def payload_nbytes(self, payloadsize):
        self['BLOCSIZE'] = payloadsize

    @property
    def frame_nbytes(self):
        """Size of the frame in bytes."""
        return self.nbytes + self.payload_nbytes

    @frame_nbytes.setter
    def frame_nbytes(self, frame_nbytes):
        self.payload_nbytes = frame_nbytes - self.nbytes

    @property
    def bps(self):
        """Bits per elementary sample."""
        return self['NBITS']

    @bps.setter
    def bps(self, bps):
        self['NBITS'] = bps

    @property
    def complex_data(self):
        """Whether the data are complex."""
        return self['OBSNCHAN'] != 1

    @property
    def npol(self):
        """Number of polarisations."""
        return self['NPOL'] // (2 if self.complex_data else 1)

    @npol.setter
    def npol(self, npol):
        self['NPOL'] = npol * (2 if self.complex_data else 1)

    @property
    def nchan(self):
        """Number of channels."""
        return self['OBSNCHAN']

    @nchan.setter
    def nchan(self, nchan):
        self['OBSNCHAN'] = operator.index(nchan)

    @property
    def sample_shape(self):
        """Shape of a sample in the payload (npol, nchan)."""
        return self.npol, self.nchan

    @sample_shape.setter
    def sample_shape(self, sample_shape):
        # Need to set nchan first to properly set npol, since nchan is
        # connected to complex_data.
        self.nchan = sample_shape[1]
        self.npol = sample_shape[0]

    @property
    def _bpcs(self):
        """Bits per complete sample."""
        # NPOL includes factor of 2 for real/complex components.
        return self['OBSNCHAN'] * self['NPOL'] * self.bps

    @property
    def sample_rate(self):
        """Number of complete samples per second.

        Can be set with a negative quantity to set `sideband`.  Overlap samples
        are not included in the rate.
        """
        return (1. / self['TBIN']) * u.Hz

    @sample_rate.setter
    def sample_rate(self, sample_rate):
        self['TBIN'] = 1. / abs(sample_rate.to_value(u.Hz))
        bw = (sample_rate.to_value(u.MHz) * self['OBSNCHAN']
              / (1 if self.complex_data else 2))
        self['OBSBW'] = bw

    @property
    def sideband(self):
        """True if upper sideband."""
        return self['OBSBW'] > 0

    @sideband.setter
    def sideband(self, sideband):
        self['OBSBW'] = (1 if sideband else -1) * abs(self['OBSBW'])

    @property
    def channels_first(self):
        """True if encoded payload ordering is (nchan, nsample, npol)."""
        # Called ``time-ordered`` in DSPSR.
        return self['PKTFMT'] != 'SIMPLE'

    @channels_first.setter
    def channels_first(self, channels_first):
        self['PKTFMT'] = '1SFA' if bool(channels_first) else 'SIMPLE'

    @property
    def samples_per_frame(self):
        """Number of complete samples in the frame, including overlap."""
        return self.payload_nbytes * 8 // self._bpcs

    @samples_per_frame.setter
    def samples_per_frame(self, samples_per_frame):
        old_payload_nbytes = self.payload_nbytes
        self.payload_nbytes = (samples_per_frame * self._bpcs + 7) // 8
        if self.samples_per_frame != samples_per_frame:
            exc = ValueError("header cannot store {} samples per frame. "
                             "Nearest is {}.".format(samples_per_frame,
                                                     self.samples_per_frame))
            self.payload_nbytes = old_payload_nbytes
            raise exc

    @property
    def overlap(self):
        """Number of complete samples that overlap with the next frame."""
        return self['OVERLAP']

    @overlap.setter
    def overlap(self, overlap):
        self['OVERLAP'] = operator.index(overlap)

    @property
    def offset(self):
        """Offset from start of observation in units of time."""
        # PKTIDX only counts valid packets, not overlap ones.
        return self['STT_OFFS'] + ((self['PKTIDX'] * self['PKTSIZE'] * 8
                                    // self._bpcs) * self['TBIN'] * u.s)

    @offset.setter
    def offset(self, offset):
        self['PKTIDX'] = int((offset / (self['TBIN'] * u.s) / self['PKTSIZE']
                              * ((self._bpcs + 7) // 8)).to(u.one).round())

    @property
    def start_time(self):
        """Start time of the observation."""
        return (Time(self['STT_IMJD'], scale='utc', format='mjd')
                + self['STT_SMJD'] * u.s)

    @start_time.setter
    def start_time(self, start_time):
        start_time = Time(start_time, scale='utc', format='isot', precision=9)
        self['STT_IMJD'] = int(start_time.mjd)
        self['STT_SMJD'] = int(np.around(
            (start_time - Time(self['STT_IMJD'], format='mjd',
                               scale=start_time.scale)).sec))

    @property
    def time(self):
        """Start time of the part of the observation covered by this header."""
        return self.start_time + self.offset

    @time.setter
    def time(self, time):
        """Set header time.

        If ``start_time`` is not already set, this sets it using the time and
        ``offset``.  Otherwise, this sets ``offset`` using the time and
        ``start_time``.

        Parameters
        ----------
        time : `~astropy.time.Time`
            Time for the first sample associated with this header.
        """
        if 'STT_IMJD' not in self.keys():
            self.start_time = time - self.offset
        else:
            self.offset = time - self.start_time

    def _ipython_key_completions_(self):
        # Enables tab-completion of header keys in IPython.
        return self.keys()

    def __eq__(self, other):
        """Whether headers have the same keys with the same values."""
        return all(self.get(k, None) == other.get(k, None)
                   for k in (set(self.keys()) | set(other.keys())))

    def __repr__(self):
        name = self.__class__.__name__
        vals = super().__repr__()
        return('<{0} {1}>'.format(name, ("\n  " + len(name) * " ").join(
            [v.rstrip() for v in vals.split('\n')])))



# Licensed under the GPLv3 - see LICENSE
import re

import astropy.units as u
from astropy.utils import lazyproperty

from ..helpers import sequentialfile as sf
from ..base.base import (
    FileBase,
    StreamReaderBase, StreamWriterBase,
    FileOpener, FileInfo)
from .header import GUPPIHeader
from .payload import GUPPIPayload
from .frame import GUPPIFrame
from .file_info import GUPPIFileReaderInfo


__all__ = ['GUPPIFileNameSequencer', 'GUPPIFileReader', 'GUPPIFileWriter',
           'GUPPIStreamBase', 'GUPPIStreamReader', 'GUPPIStreamWriter',
           'open', 'info']


class GUPPIFileNameSequencer(sf.FileNameSequencer):
    """List-like generator of GUPPI filenames using a template.

    The template is formatted, filling in any items in curly brackets with
    values from the header, as well as possibly a file number equal to the
    indexing value, indicated with '{file_nr}'.

    The length of the instance will be the number of files that exist that
    match the template for increasing values of the file number (when writing,
    it is the number of files that have so far been generated).

    Parameters
    ----------
    template : str
        Template to format to get specific filenames.  Curly bracket item
        keywords are not case-sensitive.
    header : dict-like
        Structure holding key'd values that are used to fill in the format.
        Keys must be in all caps (eg. ``DATE``), as with GUPPI header keys.

    Examples
    --------

    >>> from baseband import guppi
    >>> gfs = guppi.GUPPIFileNameSequencer(
    ...     '{date}_{file_nr:03d}.raw', {'DATE': "2018-01-01"})
    >>> gfs[10]
    '2018-01-01_010.raw'
    >>> from baseband.data import SAMPLE_PUPPI
    >>> with open(SAMPLE_PUPPI, 'rb') as fh:
    ...     header = guppi.GUPPIHeader.fromfile(fh)
    >>> template = 'puppi_{stt_imjd}_{src_name}_{scannum}.{file_nr:04d}.raw'
    >>> gfs = guppi.GUPPIFileNameSequencer(template, header)
    >>> gfs[0]
    'puppi_58132_J1810+1744_2176.0000.raw'
    >>> gfs[10]
    'puppi_58132_J1810+1744_2176.0010.raw'
    """

    def __init__(self, template, header={}):
        self.items = {}

        def check_and_convert(x):
            string = x.group().upper()
            key = string[1:-1]
            if key != 'FILE_NR':
                self.items[key] = header[key]
            return string

        # This converts template names to upper case, since header keywords are
        # all upper case.
        self.template = re.sub(r'{\w+[}:]', check_and_convert, template)

    def _process_items(self, file_nr):
        super()._process_items(file_nr)
        # Pop file_nr, as we need to capitalize it.
        self.items['FILE_NR'] = self.items.pop('file_nr')


class GUPPIFileReader(FileBase):
    """Simple reader for GUPPI files.

    Wraps a binary filehandle, providing methods to help interpret the data,
    such as `read_frame` and `get_frame_rate`. By default, frame payloads
    are mapped rather than fully read into physical memory.

    Parameters
    ----------
    fh_raw : filehandle
        Filehandle of the raw binary data file.
    """
    info = GUPPIFileReaderInfo()

    def read_header(self):
        """Read a single header from the file.

        Returns
        -------
        header : `~baseband.guppi.GUPPIHeader`
        """
        return GUPPIHeader.fromfile(self.fh_raw)

    def read_frame(self, memmap=True, verify=True):
        """Read the frame header and read or map the corresponding payload.

        Parameters
        ----------
        memmap : bool, optional
            If `True` (default), map the payload using `~numpy.memmap`, so that
            parts are only loaded into memory as needed to access data.
        verify : bool, optional
            Whether to do basic checks of frame integrity.  Default: `True`.

        Returns
        -------
        frame : `~baseband.guppi.GUPPIFrame`
            With ``.header`` and ``.payload`` properties.  The ``.data``
            property returns all data encoded in the frame.  Since this may
            be too large to fit in memory, it may be better to access the
            parts of interest by slicing the frame.
        """
        return GUPPIFrame.fromfile(self.fh_raw, memmap=memmap, verify=verify)

    def get_frame_rate(self):
        """Determine the number of frames per second.

        The routine uses the sample rate and number of samples per frame
        (excluding overlap) from the first header in the file.

        Returns
        -------
        frame_rate : `~astropy.units.Quantity`
            Frames per second.
        """
        with self.temporary_offset(0):
            header = self.read_header()
        return (header.sample_rate
                / (header.samples_per_frame - header.overlap)).to(u.Hz)


class GUPPIFileWriter(FileBase):
    """Simple writer/mapper for GUPPI files.

    Adds `write_frame` and `memmap_frame` methods to the binary file wrapper.
    The latter allows one to encode data in pieces, writing to disk as needed.
    """

    def write_frame(self, data, header=None, **kwargs):
        """Write a single frame (header plus payload).

        Parameters
        ----------
        data : `~numpy.ndarray` or `~baseband.guppi.GUPPIFrame`
            If an array, a ``header`` should be given, which will be used to
            get the information needed to encode the array, and to construct
            the GUPPI frame.
        header : `~baseband.guppi.GUPPIHeader`
            Can instead give keyword arguments to construct a header.  Ignored
            if ``data`` is a `~baseband.guppi.GUPPIFrame` instance.
        **kwargs
            If ``header`` is not given, these are used to initialize one.
        """
        if not isinstance(data, GUPPIFrame):
            data = GUPPIFrame.fromdata(data, header, **kwargs)
        return data.tofile(self.fh_raw)

    def memmap_frame(self, header=None, **kwargs):
        """Get frame by writing the header to disk and mapping its payload.

        The header is written to disk immediately, but the payload is mapped,
        so that it can be filled in pieces, by setting slices of the frame.

        Parameters
        ----------
        header : `~baseband.guppi.GUPPIHeader`
            Written to disk immediately.  Can instead give keyword arguments to
            construct a header.
        **kwargs
            If ``header`` is not given, these are used to initialize one.

        Returns
        -------
        frame: `~baseband.guppi.GUPPIFrame`
            By assigning slices to data, the payload can be encoded piecewise.
        """
        if header is None:
            header = GUPPIHeader.fromvalues(**kwargs)
        header.tofile(self.fh_raw)
        payload = GUPPIPayload.fromfile(self.fh_raw, memmap=True,
                                        header=header)
        return GUPPIFrame(header, payload)


class GUPPIStreamBase:
    """Provides samples_per_frame and fast index getting/setting."""

    _sample_shape_maker = GUPPIPayload._sample_shape_maker

    # Overriding instead of passing on a different samples_per_frame
    # in __init__ so the docstring indicates the exclusion of the overlap.
    @property
    def samples_per_frame(self):
        """"Number of complete samples per frame, excluding overlap."""
        return self.header0.samples_per_frame - self.header0.overlap

    @lazyproperty
    def _packets_per_frame(self):
        # GUPPI headers report their offsets using 'PKTIDX', the number of
        # unique UDP data packets (i.e. excluding overlap) written since the
        # start of observation.  'PKTSIZE' is the packet size in bytes.  Here
        # we calculate the packets per frame.
        return ((self.header0.payload_nbytes
                 - self.header0.overlap * self.header0._bpcs // 8)
                // self.header0['PKTSIZE'])

    def _get_index(self, header):
        # Override to avoid calculating index from time.
        return int(round((header['PKTIDX'] - self.header0['PKTIDX'])
                         / self._packets_per_frame))

    def _set_index(self, header, index):
        header.update(pktidx=self.header0['PKTIDX']
                      + index * self._packets_per_frame)


class GUPPIStreamReader(GUPPIStreamBase, StreamReaderBase):
    """GUPPI format reader.

    Allows access to GUPPI files as a continuous series of samples.

    Parameters
    ----------
    fh_raw : filehandle
        Filehandle of the raw GUPPI stream.
    squeeze : bool, optional
        If `True` (default), remove any dimensions of length unity from
        decoded data.
    subset : indexing object or tuple of objects, optional
        Specific components of the complete sample to decode (after possibly
        squeezing).  If a single indexing object is passed, it selects
        polarizations.  With a tuple, the first selects polarizations and the
        second selects channels.  If the tuple is empty (default), all
        components are read.
    verify : bool, optional
        Whether to do basic checks of frame integrity when reading.  The first
        frame of the stream is always checked, so ``verify`` is effective only
        when reading sequences of files.  Default: `True`.
    """

    def __init__(self, fh_raw, squeeze=True, subset=(), verify=True):
        fh_raw = GUPPIFileReader(fh_raw)
        header0 = fh_raw.read_header()
        super().__init__(fh_raw, header0, squeeze=squeeze,
                         subset=subset, verify=verify)

    @lazyproperty
    def stop_time(self):
        """Time at the end of the file, just after the last sample.

        This includes any overlap in the last frame.

        See also `start_time` for the start time of the file, and `time` for
        the time of the sample pointer's current offset.
        """
        return (self._get_time(self._last_header)
                + (self._last_header.samples_per_frame / self.sample_rate))

    def _get_frame(self, offset):
        normal_end = self._nsample-self._last_header.overlap
        if normal_end <= offset < self._nsample:
            # In overlap of last frame.
            frame, sample_offset = super()._get_frame(normal_end-1)
            return frame, sample_offset + 1 + offset - normal_end

        else:
            return super()._get_frame(offset)


class GUPPIStreamWriter(GUPPIStreamBase, StreamWriterBase):
    """GUPPI format writer.

    Encodes and writes sequences of samples to file.

    Parameters
    ----------
    raw : filehandle
        For writing the header and raw data to storage.
    header0 : :class:`~baseband.guppi.GUPPIHeader`
        Header for the first frame, holding time information, etc.
    squeeze : bool, optional
        If `True` (default), `write` accepts squeezed arrays as input,
        and adds any dimensions of length unity.
    """

    def __init__(self, fh_raw, header0, squeeze=True):
        assert header0.get('OVERLAP', 0) == 0, ("overlap must be 0 when "
                                                "writing GUPPI files.")
        fh_raw = GUPPIFileWriter(fh_raw)
        super().__init__(fh_raw, header0, squeeze=squeeze)

    def _make_frame(self, index):
        header = self.header0.copy()
        self._set_index(header, index)
        return self.fh_raw.memmap_frame(header)

    def _fh_raw_write_frame(self, frame):
        assert frame is self._frame
        del self._frame


class GUPPIFileOpener(FileOpener):
    FileNameSequencer = GUPPIFileNameSequencer
    non_header_keys = FileOpener.non_header_keys | {'frames_per_file'}

    def get_fh(self, name, mode, kwargs):
        if mode == 'ws' and self.is_sequence(name):
            kwargs.setdefault('file_size',
                              kwargs.pop('frames_per_file', 128)
                              * kwargs['header0'].frame_nbytes)

        return super().get_fh(name, mode, kwargs)


open = GUPPIFileOpener.create(globals(), doc="""
--- For reading a stream : (see `~baseband.guppi.base.GUPPIStreamReader`)

squeeze : bool, optional
    If `True` (default), remove any dimensions of length unity from
    decoded data.
subset : indexing object or tuple of objects, optional
    Specific components of the complete sample to decode (after possibly
    squeezing).  If a single indexing object is passed, it selects
    polarizations.  With a tuple, the first selects polarizations and the
    second selects channels.  If the tuple is empty (default), all
    components are read.

--- For writing a stream : (see `~baseband.guppi.base.GUPPIStreamWriter`)

header0 : `~baseband.guppi.GUPPIHeader`
    Header for the first frame, holding time information, etc.  Can instead
    give keyword arguments to construct a header (see ``**kwargs``).
squeeze : bool, optional
    If `True` (default), writer accepts squeezed arrays as input, and adds
    any dimensions of length unity.
frames_per_file : int, optional
    When writing to a sequence of files, sets the number of frames
    within each file.  Default: 128.
**kwargs
    If no header is given, an attempt is made to construct one from these.
    For a standard header, this would include the following.

--- Header keywords : (see :meth:`~baseband.guppi.GUPPIHeader.fromvalues`)

time : `~astropy.time.Time`
    Start time of the file.  Must have an integer number of seconds.
sample_rate : `~astropy.units.Quantity`
    Number of complete samples per second, i.e. the rate at which each
    channel of each polarization is sampled.
samples_per_frame : int
    Number of complete samples per frame.  Can alternatively give
    ``payload_nbytes``.
payload_nbytes : int
    Number of bytes per payload.  Can alternatively give ``samples_per_frame``.
offset : `~astropy.units.Quantity` or `~astropy.time.TimeDelta`, optional
    Time offset from the start of the whole observation (default: 0).
npol : int, optional
    Number of polarizations (default: 1).
nchan : int, optional
    Number of channels (default: 1).  For GUPPI, complex data is only allowed
    when nchan > 1.
bps : int, optional
    Bits per elementary sample, i.e. per real or imaginary component for
    complex data (default: 8).

Returns
-------
Filehandle
    :class:`~baseband.guppi.base.GUPPIFileReader` or
    :class:`~baseband.guppi.base.GUPPIFileWriter` (binary), or
    :class:`~baseband.guppi.base.GUPPIStreamReader` or
    :class:`~baseband.guppi.base.GUPPIStreamWriter` (stream).

Notes
-----
For streams, one can also pass to ``name`` a list of files, or a template
string that can be formatted using 'stt_imjd', 'src_name', and other header
keywords (by `~baseband.guppi.GUPPIFileNameSequencer`).

For writing, one can mimic, for example, what is done at Arecibo by using
the template 'puppi_{stt_imjd}_{src_name}_{scannum}.{file_nr:04d}.raw'.  GUPPI
typically has 128 frames per file; to change this, use the ``frames_per_file``
keyword.  ``file_size`` is set by ``frames_per_file`` and cannot be passed.

For reading, to read series such as the above, you will need to use something
like 'puppi_58132_J1810+1744_2176.{file_nr:04d}.raw'.  Here we have to pass in
the MJD, source name and scan number explicitly, since the template is used to
get the first file name, before any header is read, and therefore the only
keyword available is 'file_nr', which is assumed to be zero for the first file.
To avoid this restriction, pass in keyword arguments with values appropriate
for the first file.

One may also pass in a `~baseband.helpers.sequentialfile` object
(opened in 'rb' mode for reading or 'w+b' for writing), though for typical use
cases it is practically identical to passing in a list or template.
""")


info = FileInfo.create(globals())



from ..base.file_info import (  # noqa
    info_item,
    InfoMeta as VLBIInfoMeta,
    InfoBase as VLBIInfoBase,
    FileReaderInfo as VLBIFileReaderInfo,
    StreamReaderInfo as VLBIStreamReaderInfo)



# Licensed under the GPLv3 - see LICENSE
import numpy as np
import astropy.units as u
from astropy.utils import lazyproperty

from ..base.base import (
    FileBase, VLBIFileReaderBase,
    VLBIStreamReaderBase, StreamWriterBase,
    FileOpener, FileInfo)
from .header import Mark5BHeader
from .payload import Mark5BPayload
from .frame import Mark5BFrame
from .file_info import Mark5BFileReaderInfo


__all__ = ['Mark5BFileReader', 'Mark5BFileWriter',
           'Mark5BStreamBase', 'Mark5BStreamReader', 'Mark5BStreamWriter',
           'open', 'info']


class Mark5BFileReader(VLBIFileReaderBase):
    """Simple reader for Mark 5B files.

    Wraps a binary filehandle, providing methods to help interpret the data,
    such as `read_frame` and `get_frame_rate`.

    Parameters
    ----------
    fh_raw : filehandle
        Filehandle of the raw binary data file.
    kday : int or None
        Explicit thousands of MJD of the observation time.  Can instead
        pass an approximate ``ref_time``.
    ref_time : `~astropy.time.Time` or None
        Reference time within 500 days of the observation time, used to
        infer the full MJD.  Used only if ``kday`` is not given.
    nchan : int, optional
        Number of channels.   Default: 1.
    bps : int, optional
        Bits per elementary sample.  Default: 2.
    """

    def __init__(self, fh_raw, kday=None, ref_time=None, nchan=None, bps=2):
        self.kday = kday
        self.ref_time = ref_time
        self.nchan = nchan
        self.bps = bps
        super().__init__(fh_raw)

    def __repr__(self):
        return ("{name}(fh_raw={s.fh_raw}, kday={s.kday}, "
                "ref_time={s.ref_time}, nchan={s.nchan}, bps={s.bps})"
                .format(name=self.__class__.__name__, s=self))

    info = Mark5BFileReaderInfo()

    def read_header(self):
        """Read a single header from the file.

        Returns
        -------
        header : `~baseband.mark5b.Mark5BHeader`
        """
        return Mark5BHeader.fromfile(self, kday=self.kday,
                                     ref_time=self.ref_time)

    def read_frame(self, verify=True):
        """Read a single frame (header plus payload).

        Returns
        -------
        frame : `~baseband.mark5b.Mark5BFrame`
            With ``header`` and ``data`` properties that return the
            `~baseband.mark5b.Mark5BHeader` and data encoded in the frame,
            respectively.
        verify : bool, optional
            Whether to do basic checks of frame integrity.  Default: `True`.
        """
        if self.nchan is None:
            raise TypeError("In order to read frames, the file handle should "
                            "be initialized with nchan set.")
        return Mark5BFrame.fromfile(self.fh_raw, kday=self.kday,
                                    ref_time=self.ref_time,
                                    sample_shape=(self.nchan,),
                                    bps=self.bps, verify=verify)

    def get_frame_rate(self):
        """Determine the number of frames per second.

        This method first tries to determine the frame rate by looking for
        the highest frame number in the first second of data.  If that fails,
        it uses the time difference between two consecutive frames. This can
        fail if the headers do not store fractional seconds, or if the data
        rate is above 512 Mbps.

        Returns
        -------
        frame_rate : `~astropy.units.Quantity`
            Frames per second.
        """
        try:
            return super().get_frame_rate()
        except Exception as exc:
            with self.temporary_offset(0):
                try:
                    header0 = self.read_header()
                    self.seek(header0.payload_nbytes, 1)
                    header1 = self.read_header()
                    tdelta = header1.fraction - header0.fraction
                    if tdelta == 0.:
                        exc.args += ("frame rate can also not be determined "
                                     "from the first two headers, as they "
                                     "have identical fractional seconds.",)
                    return u.Quantity(1 / tdelta, u.Hz).round()
                except Exception:
                    pass
            raise exc

    def locate_frames(self, pattern=None, **kwargs):
        """Use a pattern to locate frame starts near the current position.

        Note that the current position is always included.

        Parameters are as for
        `baseband.base.base.VLBIFileReaderBase.locate_frames`
        except that by default the Mark 5B sync pattern is used.
        """
        if pattern is None:
            pattern = Mark5BHeader
        return super().locate_frames(pattern, **kwargs)


class Mark5BFileWriter(FileBase):
    """Simple writer for Mark 5B files.

    Adds `write_frame` method to the VLBI binary file wrapper.
    """

    def write_frame(self, data, header=None, bps=2, valid=True, **kwargs):
        """Write a single frame (header plus payload).

        Parameters
        ----------
        data : `~numpy.ndarray` or :`~baseband.mark5b.Mark5BFrame`
            If an array, ``header`` should be given, which will be used to
            get the information needed to encode the array, and to construct
            the Mark 5B frame.
        header : `~baseband.mark5b.Mark5BHeader`
            Can instead give keyword arguments to construct a header.  Ignored
            if ``data`` is a `~baseband.mark5b.Mark5BFrame` instance.
        bps : int, optional
            Bits per elementary sample, to use when encoding the payload.
            Ignored if ``data`` is a `~baseband.mark5b.Mark5BFrame` instance.
            Default: 2.
        valid : bool, optional
            Whether the data are valid; if `False`, a payload filled with an
            appropriate pattern will be crated.  Ignored if ``data`` is a
            `~baseband.mark5b.Mark5BFrame` instance.  Default: `True`.
        **kwargs
            If ``header`` is not given, these are used to initialize one.
        """
        if not isinstance(data, Mark5BFrame):
            data = Mark5BFrame.fromdata(data, header, bps=bps, valid=valid,
                                        **kwargs)
        return data.tofile(self.fh_raw)


class Mark5BStreamBase:
    """Provides sample shape maker and fast time and index getting/setting."""

    _sample_shape_maker = Mark5BPayload._sample_shape_maker

    def _get_time(self, header):
        return header.get_time(frame_rate=self._frame_rate)

    def _set_time(self, header, time):
        # Pass frame_rate, which is needed to calculate the frame number,
        # and use update so that also the CRC is updated.
        header.update(time=time, frame_rate=self._frame_rate)

    def _get_index(self, header):
        # Override to provide index faster, without calculating times.
        # TODO: OK to ignore leap seconds? Not sure what writer does.
        return int(round(self._frame_rate.to_value(u.Hz)
                         * (header.seconds - self.header0.seconds
                            + 86400 * (header.kday - self.header0.kday
                                       + header.jday - self.header0.jday))
                         + header['frame_nr'] - self.header0['frame_nr']))

    def _set_index(self, header, index):
        # Override to avoid explicit time calculations.
        dt, frame_nr = divmod(index + self.header0['frame_nr'],
                              int(round(self._frame_rate.to_value(u.Hz))))
        fraction = (frame_nr/self._frame_rate).to_value(u.s)
        dd, seconds = divmod(dt + self.header0.seconds, 86400)
        dk, jday = divmod(dd + self.header0.jday, 1000)
        kday = dk*1000 + self.header0.kday
        # Use update to ensure CRC is updated as well.
        header.update(frame_nr=frame_nr, fraction=fraction,
                      seconds=seconds, jday=jday, kday=kday)


class Mark5BStreamReader(Mark5BStreamBase, VLBIStreamReaderBase):
    """VLBI Mark 5B format reader.

    Allows access a Mark 5B file as a continues series of samples.

    Parameters
    ----------
    fh_raw : filehandle
        Filehandle of the raw Mark 5B stream.
    sample_rate : `~astropy.units.Quantity`, optional
        Number of complete samples per second, i.e. the rate at which each
        channel is sampled.  If `None` (default), will be inferred from
        scanning one second of the file or, failing that, using the time
        difference between two consecutive frames.
    kday : int or None
        Explicit thousands of MJD of the observation start time (eg. ``57000``
        for MJD 57999), used to infer the full MJD from the header's time
        information.  Can instead pass an approximate ``ref_time``.
    ref_time : `~astropy.time.Time` or None
        Reference time within 500 days of the observation start time, used
        to infer the full MJD.  Only used if ``kday`` is not given.
    nchan : int
        Number of channels.  Needs to be explicitly passed in.
    bps : int, optional
        Bits per elementary sample.  Default: 2.
    squeeze : bool, optional
        If `True` (default), remove any dimensions of length unity from
        decoded data.
    subset : indexing object, optional
        Specific channels of the complete sample to decode (after possible
        squeezing). If an empty tuple (default), all channels are read.
    fill_value : float or complex
        Value to use for invalid or missing data. Default: 0.
    verify : bool or 'fix', optional
        Whether to do basic checks of frame integrity when reading.
        Default: 'fix', which implies basic verification and replacement
        of gaps with zeros.
    """

    def __init__(self, fh_raw, sample_rate=None, kday=None, ref_time=None,
                 nchan=None, bps=2, squeeze=True, subset=(), fill_value=0.,
                 verify='fix'):

        if nchan is None:
            raise TypeError("Mark 5B stream reader requires nchan to be "
                            "explicity passed in.")

        if kday is None and ref_time is None:
            raise TypeError("Mark 5B stream reader requires either kday or "
                            "ref_time to be passed in.")

        fh_raw = Mark5BFileReader(fh_raw, nchan=nchan, bps=bps,
                                  ref_time=ref_time, kday=kday)
        header0 = fh_raw.find_header()
        super().__init__(
            fh_raw, header0, sample_rate=sample_rate,
            samples_per_frame=header0.payload_nbytes * 8 // bps // nchan,
            sample_shape=(nchan,), bps=bps,
            squeeze=squeeze, subset=subset, fill_value=fill_value,
            verify=verify)
        # Use ref_time in preference to kday so we can handle files that
        # span a change in 1000s of MJD.
        self.fh_raw.kday = None
        self.fh_raw.ref_time = self.start_time

    @lazyproperty
    def _last_header(self):
        """Last header of the file."""
        last_header = super()._last_header
        # Infer kday, assuming the end of the file is no more than
        # 500 days away from the start.
        last_header.infer_kday(self.start_time)
        return last_header


class Mark5BStreamWriter(Mark5BStreamBase, StreamWriterBase):
    """VLBI Mark 5B format writer.

    Encodes and writes sequences of samples to file.

    Parameters
    ----------
    fh_raw : filehandle
        For writing filled sets of frames to storage.
    header0 : `~baseband.mark5b.Mark5BHeader`
        Header for the first frame, holding time information, etc.
    sample_rate : `~astropy.units.Quantity`
        Number of complete samples per second, i.e. the rate at which each
        channel is sampled.  Needed to calculate header timestamps.
    nchan : int, optional
        Number of channels.  Default: 1.
    bps : int, optional
        Bits per elementary sample.  Default: 2.
    squeeze : bool, optional
        If `True` (default), `write` accepts squeezed arrays as input, and
        adds any dimensions of length unity.
    """

    def __init__(self, fh_raw, header0=None, sample_rate=None, nchan=1, bps=2,
                 squeeze=True):
        fh_raw = Mark5BFileWriter(fh_raw)
        super().__init__(
            fh_raw, header0, sample_rate=sample_rate,
            samples_per_frame=header0.payload_nbytes * 8 // bps // nchan,
            sample_shape=(nchan,), bps=bps, squeeze=squeeze)
        # Initial frame, reused for every other one.
        self._frame = Mark5BFrame.fromdata(
            np.zeros((self.samples_per_frame, nchan)),
            header0.copy(), bps=bps)


class Mark5BFileOpener(FileOpener):
    def get_header0(self, kwargs):
        if ('time' in kwargs and 'sample_rate' in kwargs
                and 'frame_rate' not in kwargs):
            bps = kwargs.get('bps', 2)
            nchan = kwargs.get('nchan', 1)
            samples_per_frame = Mark5BHeader.payload_nbytes * 8 // bps // nchan
            kwargs['frame_rate'] = kwargs['sample_rate'] / samples_per_frame
        header0 = super().get_header0(kwargs)
        kwargs.pop('frame_rate', None)
        return header0


open = Mark5BFileOpener.create(globals(), doc="""
--- For reading a stream : (see `~baseband.mark5b.base.Mark5BStreamReader`)

sample_rate : `~astropy.units.Quantity`, optional
    Number of complete samples per second, i.e. the rate at which each channel
    is sampled.  If `None` (default), will be inferred from scanning one
    second of the file or, failing that, using the time difference between two
    consecutive frames.
kday : int or None
    Explicit thousands of MJD of the observation start time (eg. ``57000`` for
    MJD 57999), used to infer the full MJD from the header's time information.
    Can instead pass an approximate ``ref_time``.
ref_time : `~astropy.time.Time` or None
    Reference time within 500 days of the observation start time, used to infer
    the full MJD.  Only used if ``kday`` is not given.
nchan : int, optional
    Number of channels.  Default: 1.
bps : int, optional
    Bits per elementary sample.  Default: 2.
squeeze : bool, optional
    If `True` (default), remove any dimensions of length unity from
    decoded data.
subset : indexing object, optional
    Specific channels of the complete sample to decode (after possible
    squeezing). If an empty tuple (default), all channels are read.
fill_value : float or complex
    Value to use for invalid or missing data. Default: 0.
verify : bool or 'fix', optional
    Whether to do basic checks of frame integrity when reading.
    Default: 'fix', which implies basic verification and replacement
    of gaps with zeros.

--- For writing a stream : (see `~baseband.mark5b.base.Mark5BStreamWriter`)

header0 : :class:`~baseband.mark5b.Mark5BHeader`
    Header for the first frame, holding time information, etc.  Can instead
    give keyword arguments to construct a header (see ``**kwargs``).
sample_rate : `~astropy.units.Quantity`
    Number of complete samples per second, i.e. the rate at which each
    channel is sampled.  Needed to calculate header timestamps.
nchan : int, optional
    Number of channels.  Default: 1.
bps : int, optional
    Bits per elementary sample.  Default: 2.
squeeze : bool, optional
    If `True` (default), writer accepts squeezed arrays as input,
    and adds channel and thread dimensions if they have length unity.
file_size : int or None, optional
    When writing to a sequence of files, the maximum size of one file in bytes.
    If `None` (default), the file size is unlimited, and only the first
    file will be written to.
**kwargs
    If no header is given, an attempt is made to construct one from these.
    For a standard header, the following suffices.

--- Header kwargs : (see :meth:`~baseband.mark5b.Mark5BHeader.fromvalues`)

time : `~astropy.time.Time`
    Start time of the file.  Sets bcd-encoded unit day, hour, minute,
    second, and fraction, as well as the frame number, in the header.

Returns
-------
Filehandle
    :class:`~baseband.mark5b.base.Mark5BFileReader` or
    :class:`~baseband.mark5b.base.Mark5BFileWriter` (binary), or
    :class:`~baseband.mark5b.base.Mark5BStreamReader` or
    :class:`~baseband.mark5b.base.Mark5BStreamWriter` (stream).

Notes
-----
One can also pass to ``name`` a list, tuple, or subclass of
`~baseband.helpers.sequentialfile.FileNameSequencer`.  For writing to multiple
files, the ``file_size`` keyword must be passed or only the first file will be
written to.  One may also pass in a `~baseband.helpers.sequentialfile` object
(opened in 'rb' mode for reading or 'w+b' for writing), though for typical use
cases it is practically identical to passing in a list or template.
""")


info = FileInfo.create(globals())



# Licensed under the GPLv3 - see LICENSE
"""
Base definitions for baseband frames.

Defines a frame class FrameBase that can be used to hold a header and a
payload, providing access to the values encoded in both.
"""
import numpy as np


__all__ = ['FrameBase']


class FrameBase:
    """Representation of a baseaband frame, consisting of a header and payload.

    Parameters
    ----------
    header : ``cls._header_class``
        Wrapper around the header, providing mechanisms to decode it.
    payload : ``cls._payload_class``
        Wrapper around the payload, providing mechanisms to decode it.
    valid : bool, optional
        Whether the data are valid.  Default: class default, `True`.
    verify : bool, optional
        Whether to do basic verification of integrity.  Default: `True`.

    Notes
    -----
    The Frame can also be instantiated using class methods:

      fromfile : read header and payload from a filehandle

      fromdata : encode data as payload

    Of course, one can also do the opposite:

      tofile : method to write header and payload to filehandle

      data : property that yields full decoded payload

    One can decode part of the payload by indexing or slicing the frame.
    If the frame does not contain valid data, all values returned are set
    to ``self.fill_value``.

    A number of properties are defined: `shape` and `dtype` are the shape
    and type of the data array, and `nbytes` the frame size in bytes.
    Furthermore, the frame acts as a dictionary, with keys those of the header.
    Any attribute that is not defined on the frame itself, such as ``.time``
    will be looked up on the header as well.
    """

    _header_class = None
    _payload_class = None
    _fill_value = 0.
    _valid = True

    def __init__(self, header, payload, valid=None, verify=True):
        self.header = header
        self.payload = payload
        if valid is not None:
            self.valid = valid
        if verify:
            self.verify()

    def verify(self):
        """Simple verification.  To be added to by subclasses."""
        assert isinstance(self.header, self._header_class)
        assert isinstance(self.payload, self._payload_class)
        payload_nbytes = getattr(self.header, 'payload_nbytes', None)
        if payload_nbytes is not None:
            assert self.payload.nbytes == payload_nbytes

    @property
    def valid(self):
        """Whether frame contains valid data."""
        return self._valid

    @valid.setter
    def valid(self, valid):
        self._valid = bool(valid)

    @classmethod
    def fromfile(cls, fh, memmap=None, valid=None, verify=True, **kwargs):
        """Read a frame from a filehandle.

        Parameters
        ----------
        fh : filehandle
            Handle to read the frame from
        memmap : bool, optional
            If `False`, read payload from file.  If `True`, map the payload
            in memory (see `~numpy.memmap`).  Only useful for large payloads.
            Default: as set by payload class.
        valid : bool, optional
            Whether the data are valid.  Default: inferred from header or
            payload read from file if possible, otherwise `True`.
        verify : bool, optional
            Whether to do basic verification of integrity.  Default: `True`.
        **kwargs
            Extra arguments that help to initialize the payload.
        """
        header = cls._header_class.fromfile(fh, verify=verify)
        payload = cls._payload_class.fromfile(fh, header=header,
                                              memmap=memmap, **kwargs)
        return cls(header, payload, valid=valid, verify=verify)

    def tofile(self, fh):
        """Write encoded frame to filehandle."""
        self.header.tofile(fh)
        self.payload.tofile(fh)

    @classmethod
    def fromdata(cls, data, header=None, *, valid=None, verify=True, **kwargs):
        """Construct frame from data and header.

        Parameters
        ----------
        data : `~numpy.ndarray`
            Array holding data to be encoded.
        header : header instance, optional
            Header for the frame.
        valid : bool, optional
            Whether the data are valid.  Default: inferred from header if
            possible, otherwise `True`.
        verify : bool, optional
            Whether to verify the header and frame correctness.
        **kwargs
            Used to initialize the header, if not given.
        """
        if header is None:
            header = cls._header_class.fromvalues(verify=verify, **kwargs)
        payload = cls._payload_class.fromdata(data, header=header)
        return cls(header, payload, valid=valid, verify=verify)

    @property
    def sample_shape(self):
        """Shape of a sample in the frame (nchan,)."""
        return self.payload.sample_shape

    def __len__(self):
        """Number of samples in the frame."""
        return len(self.payload)

    @property
    def shape(self):
        """Shape of the frame data."""
        return (len(self),) + self.sample_shape

    @property
    def size(self):
        """Total number of component samples in the frame data."""
        prod = 1
        for dim in self.shape:
            prod *= dim
        return prod

    @property
    def ndim(self):
        """Number of dimensions of the frame data."""
        return len(self.shape)

    @property
    def dtype(self):
        """Numeric type of the frame data."""
        return self.payload.dtype

    @property
    def nbytes(self):
        """Size of the encoded frame in bytes."""
        return self.header.nbytes + self.payload.nbytes

    @property
    def fill_value(self):
        """Value to replace invalid data in the frame."""
        return self._fill_value

    @fill_value.setter
    def fill_value(self, fill_value):
        self._fill_value = fill_value

    def __array__(self, dtype=None):
        """Interface to arrays."""
        if dtype is None or dtype == self.dtype:
            return self.data
        else:
            return self.data.astype(dtype)

    # Header behaves as a dictionary, while Payload can be indexed/sliced.
    # Let frame behave appropriately.
    def __getitem__(self, item=()):
        if isinstance(item, str):
            return self.header.__getitem__(item)
        elif self.valid:
            return self.payload.__getitem__(item)
        else:
            data_shape = np.empty(self.shape, dtype=bool)[item].shape
            return np.full(data_shape, self.fill_value,
                           dtype=self.dtype)

    data = property(__getitem__, doc="Full decoded frame.")

    def __setitem__(self, item, value):
        if isinstance(item, str):
            self.header.__setitem__(item, value)
        else:
            self.payload.__setitem__(item, value)

    def keys(self):
        return self.header.keys()

    def _ipython_key_completions_(self):
        # Enables tab-completion of header keys in IPython.
        return self.header.keys()

    def __contains__(self, key):
        return key in self.keys()

    # Try to get any attribute not on the frame from the header properties.
    def __getattr__(self, attr):
        if (attr in self.header._properties
                or attr in ('get_time', 'set_time', 'update')):
            return getattr(self.header, attr)
        else:
            # Raise appropriate error.
            return self.__getattribute__(attr)

    def __setattr__(self, attr, value):
        if (attr not in {'header', 'payload', 'valid'}
                and attr in getattr(getattr(self, 'header', None),
                                    '_properties', ())):
            return setattr(self.header, attr, value)
        else:
            return super().__setattr__(attr, value)

    # For tests, it is useful to define equality.
    def __eq__(self, other):
        return (type(self) is type(other)
                and self.valid == other.valid
                and self.header == other.header
                and self.payload == other.payload)



from ..base.base import (  # noqa
    HeaderNotFoundError,
    FileBase as _FileBase,
    VLBIFileReaderBase,
    StreamBase as _StreamBase,
    VLBIStreamReaderBase as _VLBIStreamReaderBase,
    StreamWriterBase as _StreamWriterBase,
    FileOpener as _FileOpener)


class VLBIFileBase(_FileBase):
    pass


class _UnslicedShape:
    def __init__(self, *args, **kwargs):
        if 'unsliced_shape' in kwargs:
            kwargs['sample_shape'] = kwargs.pop('unsliced_shape', None)
        super().__init__(*args, **kwargs)


class VLBIStreamBase(_UnslicedShape, _StreamBase):
    pass


class VLBIStreamReaderBase(_UnslicedShape, _VLBIStreamReaderBase):
    pass


class VLBIStreamWriterBase(_UnslicedShape, _StreamWriterBase):
    pass


def make_opener(fmt, classes, doc='', append_doc=True):
    opener = _FileOpener.create(classes, doc)
    if not append_doc:
        opener.__doc__ = doc
    return opener



# Licensed under the GPLv3 - see LICENSE
"""The GuppiFileReaderInfo property.

Overrides what can be gotten from the first header.
"""
from ..base.file_info import FileReaderInfo, info_item


__all__ = ['GUPPIFileReaderInfo']


class GUPPIFileReaderInfo(FileReaderInfo):
    # Get sample_rate from header rather than calculate it from frame_rate
    # and samples_per_frame, since we need to correct for overlap.
    attr_names = list(FileReaderInfo.attr_names)
    attr_names.insert(attr_names.index('samples_per_frame')+1, 'overlap')
    attr_names = tuple(attr_names)
    """Attributes that the container provides."""

    overlap = info_item(needs='header0', doc=(
        'Number of complete samples that overlap between frames.'))
    sample_rate = info_item(needs='header0', doc=(
        'Number of complete samples per second.'))



# Licensed under the GPLv3 - see LICENSE
"""
Definitions for GSB Headers, using the timestamp files.

Somewhat out of data description for phased data:
http://gmrt.ncra.tifr.res.in/gmrt_hpage/sub_system/gmrt_gsb/GSB_beam_timestamp_note_v1.pdf
and for rawdump data
http://gmrt.ncra.tifr.res.in/gmrt_hpage/sub_system/gmrt_gsb/GSB_rawdump_data_format_v2.pdf
"""
import os

import numpy as np
from astropy import units as u
from astropy.time import Time, TimeString
from astropy.time.formats import erfa  # astropy-version independent.

from ..base.header import HeaderParserBase, ParsedHeaderBase, ParserDict


__all__ = ['TimeGSB', 'GSBHeader', 'GSBRawdumpHeader', 'GSBPhasedHeader']


class TimeGSB(TimeString):
    """GSB header date-time format ``YYYY MM DD HH MM SS 0.SSSSSSSSS``.

    For example, ``2000 01 01 00 00 00 0.000000000`` is midnight on
    January 1, 2000.
    """
    # Implicitly uses the metaclass astropy.time.formats.TimeFormatMeta to
    # register with astropy.Time.

    name = 'gsb'

    def set_jds(self, val1, val2):
        """Parse the time strings contained in val1 and set jd1, jd2"""
        iterator = np.nditer([val1, None, None, None, None, None, None],
                             op_dtypes=([val1.dtype] + 5 * [np.intc]
                                        + [np.double]))
        try:
            for val, iy, im, id, ihr, imin, dsec in iterator:
                timestr = val.item()
                components = timestr.split()
                iy[...], im[...], id[...], ihr[...], imin[...], sec = (
                    int(component) for component in components[:-1])
                dsec[...] = sec + float(components[-1])
        except Exception:
            raise ValueError('Time {0} does not match {1} format'
                             .format(timestr, self.name))

        self.jd1, self.jd2 = erfa.dtf2d(
            self.scale.upper().encode('utf8'), *iterator.operands[1:])

    def to_value(self, parent=None):
        scale = self.scale.upper().encode('ascii'),
        iys, ims, ids, ihmsfs = erfa.d2dtf(scale, self.precision,
                                           self.jd1, self.jd2)
        ihrs = ihmsfs['h']
        imins = ihmsfs['m']
        isecs = ihmsfs['s']
        ifracs = ihmsfs['f']

        fmt = ('{0:04d} {1:02d} {2:02d} {3:02d} {4:02d} {5:02d} '
               '0.{6:0' + str(self.precision) + 'd}')
        outs = []
        for iy, im, id, ihr, imin, isec, ifracsec in np.nditer(
                [iys, ims, ids, ihrs, imins, isecs, ifracs]):
            outs.append(fmt.format(int(iy), int(im), int(id), int(ihr),
                                   int(imin), int(isec), int(ifracsec)))

        return np.array(outs).reshape(self.jd1.shape)

    value = property(to_value)


def make_parser(index, length, forward=None, backward=None, default=None):
    if length > 1:
        index = slice(index, index + length)

    def parser(items):
        return forward(items[index])

    return parser


def make_setter(index, length, forward, backward, default=None):
    def setter(items, value):
        value = backward(value)
        if length == 1:
            items[index] = value
        else:
            for i, val in enumerate(value):
                items[index + i] = val
        return items

    return setter


def get_default(index, length, forward, backward, default=None):
    return default


class GSBHeaderParser(HeaderParserBase):
    """Parser & setter for GSB timestamp keywords.

    A dictionary of header keywords, with values that describe how they are
    encoded in the GSB header.  Initialisation is as a normal dict,
    with (ordered) key, value pairs, with each value a tuple containing:

    index : int
        Index into the header words for this key.
    length : int
        Number of words included in this key.
    forward : callable
        Function to decode the data.
    backward : callable
        Function to encode the value.
    default : object or None
        Possible default value.

    Notes
    -----
    For GSB, this parsing technique is a bit of overkill, but it allows re-use
    of handy methods from the VLBI header parser.
    """

    parsers = ParserDict(make_parser)
    setters = ParserDict(make_setter)
    defaults = ParserDict(get_default)


class GSBHeader(ParsedHeaderBase):
    """GSB Header, based on a line from a timestamp file.

    Parameters
    ----------
    words : list of str, or None
        If `None`, set to a list of empty strings for later initialisation.
    mode : str or None, optional
        Mode in which data was taken: 'phased' or 'rawdump'. If `None`, it
        is determined from the words.
    nbytes : int or None, optional
        Number of characters in the header, including trailing blank spaces and
        carriage returns.  If `None`, is determined from the words assuming
        one trailing blank space and one CR.
    verify : bool, optional
        Whether to do basic verification of integrity.  Default: `True`.

    Returns
    -------
    header : `GSBHeader` subclass
        As appropriate for the mode.
    """
    _mode = None
    _gsb_header_classes = {}

    def __new__(cls, words=None, mode=None, nbytes=None, utc_offset=5.5*u.hr,
                verify=True):

        if cls is GSBHeader:
            if mode is None:
                if words is None:
                    raise TypeError("cannot construct an empty GSB header "
                                    "without knowing the mode.")

                mode = 'rawdump' if len(words) == 7 else 'phased'

            cls = cls._gsb_header_classes.get(mode)

        # We intialise GSBHeader subclasses, so their __init__ will be called.
        return super().__new__(cls)

    def __init__(self, words, mode=None, nbytes=None, utc_offset=5.5*u.hr,
                 verify=True):
        if words is None:
            words = [''] * self._number_of_words
        if mode is not None:
            self._mode = mode
        self._nbytes = nbytes
        self.utc_offset = utc_offset
        super().__init__(words, verify=verify)

    def verify(self):
        assert self.mode == self.__class__._mode
        assert len(self.words) == self._number_of_words

    @property
    def mode(self):
        """Mode in which data was taken: 'phased' or 'rawdump'."""
        return self._mode

    @property
    def nbytes(self):
        """Size of the header in characters.

        Assumes the string terminates in one blank space and one carriage
        return.
        """
        if self._nbytes is None:
            self._nbytes = len(' '.join(self.words) + os.linesep)
        return self._nbytes

    @classmethod
    def fromfile(cls, fh, *args, **kwargs):
        """Read GSB Header from a line from a timestamp file.

        Arguments are the same as for class initialisation.  The header
        constructed will be immutable.
        """
        start_pos = fh.tell()
        s = fh.readline()
        if s == '':
            raise EOFError
        nbytes = fh.tell() - start_pos
        return cls(tuple(s.split()), mode=None, nbytes=nbytes, *args, **kwargs)

    def tofile(self, fh):
        """Write GSB header as a line to the filehandle."""
        return fh.write(' '.join(self.words) + '\n')

    @classmethod
    def fromvalues(cls, mode=None, nbytes=None, *args, **kwargs):
        if mode is None and cls._mode is None:
            if set(kwargs.keys()) & {'pc', 'pc_time',
                                     'seq_nr', 'mem_block'}:
                mode = 'phased'
            else:
                raise TypeError("cannot construct a GSB header from "
                                "values without knowing the mode.")
        return super().fromvalues(mode, nbytes, *args, **kwargs)

    @classmethod
    def fromkeys(cls, mode=None, nbytes=None, *args, **kwargs):
        if mode is None and cls._mode is None:
            if set(kwargs.keys()) & {'pc', 'seq_nr', 'mem_block'}:
                mode = 'phased'
            else:
                mode = 'rawdump'
        return super().fromkeys(mode, nbytes, *args, **kwargs)

    def seek_offset(self, n, nbytes=None):
        """Offset in bytes needed to move a file pointer to another header.

        Some GSB headers have variable size and hence one cannot trivially jump
        to another entry in a timestamp file.  This routine allows one to
        calculate the offset required to move the file pointer ``n`` headers.

        Parameters
        ----------
        n : int
            The number of headers to move to, relative to the present header.
        nbytes : int, optional
            The size in bytes of the present header (if not given, will use
            the header's `nbytes` property).
        """
        if nbytes is None:
            nbytes = self.nbytes
        return n * nbytes


class GSBRawdumpHeader(GSBHeader):
    """GSB rawdump header."""

    _mode = 'rawdump'
    _number_of_words = 7
    _gps_time_precision = 9
    _properties = ('gps_time', 'time')

    _header_parser = GSBHeaderParser(
        (('gps', (0, 7, ' '.join, str.split)),))

    @property
    def gps_time(self):
        return Time(self['gps'], format='gsb',
                    precision=self._gps_time_precision) - self.utc_offset

    @gps_time.setter
    def gps_time(self, time):
        t = time + self.utc_offset
        t.precision = self._gps_time_precision
        self['gps'] = t.gsb

    time = gps_time


class GSBPhasedHeader(GSBRawdumpHeader):
    """GSB phased header."""

    _mode = 'phased'
    _number_of_words = GSBRawdumpHeader._number_of_words + 7 + 2
    _pc_time_precision = 6
    _properties = ('time', 'pc_time') + GSBRawdumpHeader._properties

    _header_parser = GSBHeaderParser(
        (('pc', (0, 7, ' '.join, str.split)),
         ('gps', (7, 7, ' '.join, str.split)),
         ('seq_nr', (14, 1, int, str, 0)),
         ('mem_block', (15, 1, int, str, 0))))

    @property
    def pc_time(self):
        return Time(self['pc'], format='gsb',
                    precision=self._pc_time_precision) - self.utc_offset

    @pc_time.setter
    def pc_time(self, time):
        t = time + self.utc_offset
        t.precision = self._pc_time_precision
        self['pc'] = t.gsb

    @property
    def time(self):
        return self.gps_time

    @time.setter
    def time(self, time):
        self.gps_time = time
        self.pc_time = time

    def seek_offset(self, n, nbytes=None):
        """Offset in bytes needed to move a file pointer to another header.

        GSB headers for phased data differ in size depending on the sequence
        number, making it impossible to trivially jump to another entry in a
        timestamp file.  This routine allows one to calculate the offset
        required to move the file pointer ``n`` headers.

        Parameters
        ----------
        n : int
            The number of headers to move to, relative to the present header.
        nbytes : int, optional
            The size in bytes of the present header (if not given, will use
            the header's `nbytes` property).
        """
        if nbytes is None:
            nbytes = self.nbytes
        # Initial guess assuming all headers have same nbytes.
        guess = n * nbytes
        # Get number of digits of current sequence number.
        seq = self['seq_nr']
        ndseq = len(str(seq))
        # Find the sequence number we're trying to reach.
        seq_sub_targ = seq + n
        # And get number of digits for this value.
        ndtarg = len(str(seq_sub_targ))
        # If numbers not the same, correct appropriately.
        while ndseq != ndtarg:
            if n > 0:
                next_power_of_ten = int('1' + ndseq * '0')
                guess += seq_sub_targ - next_power_of_ten
                ndseq += 1
            else:
                next_power_of_ten = int('1' + (ndseq - 1) * '0')
                guess += next_power_of_ten - seq_sub_targ
                ndseq -= 1

        return guess


GSBHeader._gsb_header_classes.update(rawdump=GSBRawdumpHeader,
                                     phased=GSBPhasedHeader)



# Licensed under the GPLv3 - see LICENSE
"""
Definitions for VLBI Mark 5B frames.

Implements a Mark5BFrame class that can be used to hold a header and a
payload, providing access to the values encoded in both.

For the specification, see
https://www.haystack.mit.edu/tech/vlbi/mark5/docs/Mark%205B%20users%20manual.pdf
"""
import numpy as np

from ..base.frame import FrameBase
from .header import Mark5BHeader
from .payload import Mark5BPayload


__all__ = ['Mark5BFrame']


class Mark5BFrame(FrameBase):
    """Representation of a Mark 5B frame, consisting of a header and payload.

    Parameters
    ----------
    header : `~baseband.mark5b.Mark5BHeader`
        Wrapper around the encoded header words, providing access to the
        header information.
    payload : `~baseband.mark5b.Mark5BPayload`
        Wrapper around the payload, provding mechanisms to decode it.
    valid : bool or None
        Whether the data are valid.  If `None` (default), the validity will be
        determined by checking whether the payload consists of the fill pattern
        0x11223344.
    verify : bool
        Whether to do basic verification of integrity (default: `True`)

    Notes
    -----
    The Frame can also be read instantiated using class methods:

      fromfile : read header and payload from a filehandle

      fromdata : encode data as payload

    Of course, one can also do the opposite:

      tofile : method to write header and payload to filehandle

      data : property that yields full decoded payload

    A number of properties are defined: `shape`, `dtype` and `size` are
    the shape, type and number of complete samples of the data array, and
    `nbytes` the frame size in bytes.  Furthermore, the frame acts as a
    dictionary, with keys those of the header.  Any attribute that is not
    defined on the frame itself, such as ``.time`` will be looked up on the
    header as well.
    """

    _header_class = Mark5BHeader
    _payload_class = Mark5BPayload
    _fill_pattern = 0x11223344

    def __init__(self, header, payload, valid=None, verify=True):
        if valid is None:
            # Is this payload OK?  Usually yes, so short-circuit on first few.
            valid = (payload.words[0] != self._fill_pattern
                     or payload.words[1] != self._fill_pattern
                     or payload.words[2] != self._fill_pattern
                     or (payload.words[3:] != self._fill_pattern).any())

        super().__init__(header, payload, valid, verify)

    @classmethod
    def fromfile(cls, fh, *, kday=None, ref_time=None,
                 sample_shape=(1,), bps=2, valid=None, verify=True):
        """Read a frame from a filehandle.

        Parameters
        ----------
        fh : filehandle
            To read the header and payload from.
        kday : int or None
            Explicit thousands of MJD of the observation time.  Can instead
            pass an approximate ``ref_time``.
        ref_time : `~astropy.time.Time` or None
            Reference time within 500 days of the observation time, used to
            infer the full MJD.  Used only if ``kday`` is not given.
        sample_shape : tuple of int, optional
            Can have only one entry, the number of channels.   Default: (1,).
        bps : int, optional
            Bits per elementary sample.  Default: 2.
        verify : bool
            Whether to do basic checks of frame integrity (default: `True`).
        """
        header = cls._header_class.fromfile(fh, kday=kday, ref_time=ref_time,
                                            verify=verify)
        payload = cls._payload_class.fromfile(fh, sample_shape=sample_shape,
                                              bps=bps)
        return cls(header, payload, valid, verify)

    @classmethod
    def fromdata(cls, data, header=None, bps=2, valid=True, verify=True,
                 **kwargs):
        """Construct frame from data and header.

        Parameters
        ----------
        data : `~numpy.ndarray`
            Array holding data to be encoded.
        header : `~baseband.mark5b.Mark5BHeader` or None
            If not given, will attempt to generate one using the keywords.
        bps : int
            Bits per elementary sample.  Default: 2.
        valid : bool
            Whether the data are valid (default: `True`).  If not, the payload
            will be set to a fill pattern.
        verify : bool
            Whether to do basic checks of frame integrity (default: `True`).
        """
        if header is None:
            header = Mark5BHeader.fromvalues(verify=verify, **kwargs)
        payload = cls._payload_class.fromdata(data, bps=bps)
        return cls(header, payload, valid=valid, verify=verify)

    def tofile(self, fh):
        """Write encoded frame to filehandle."""
        self.header.tofile(fh)
        if self.valid:
            self.payload.tofile(fh)
        else:
            fh.write(np.full_like(self.payload.words,
                                  self._fill_pattern).tobytes())



# Licensed under the GPLv3 - see LICENSE.rst
"""Distributed Acquisition and Data Analysis (DADA) format reader/writer."""
from .base import open, info, DADAFileNameSequencer  # noqa
from .header import DADAHeader  # noqa
from .payload import DADAPayload  # noqa
from .frame import DADAFrame  # noqa



# Licensed under the GPLv3 - see LICENSE
"""
Definitions for GSB payloads.

Implements a GSBPayload class used to store payload blocks, and decode to
or encode from a data array.

See http://gmrt.ncra.tifr.res.in/gmrt_hpage/sub_system/gmrt_gsb/index.htm
"""
from collections import namedtuple

import numpy as np

from ..base.payload import PayloadBase


__all__ = ['GSBPayload']


shift40 = np.array([4, 0], np.int8)
shift04 = np.array([0, 4], np.int8)


def decode_4bit(words):
    """Decode 4-bit data.

    For a given int8 byte containing bits 76543210,
    the first sample is in 3210, the second in 7654, and both are interpreted
    as signed 4-bit integers.
    """
    # left_shift(byte[:,np.newaxis], shift40):  [3210xxxx, 76543210]
    split = np.left_shift(words[:, np.newaxis], shift40).ravel()
    # right_shift(..., 4):                      [33333210, 77777654]
    # so least significant bits go first.
    split >>= 4
    return split.astype(np.float32)


def decode_8bit(words):
    """GSB decoder for data stored using 8 bit signed integer.
    """
    return words.astype(np.float32)


def encode_4bit(values):
    b = np.clip(np.around(values), -8, 7).astype(np.int8).reshape(-1, 2)
    b &= 0xf
    b <<= shift04
    return b[:, 0] | b[:, 1]


def encode_8bit(values):
    return np.clip(np.rint(values), -128, 127).astype(np.int8)


class GSBPayload(PayloadBase):
    """Container for decoding and encoding GSB payloads.

    Parameters
    ----------
    words : `~numpy.ndarray`
        Array containg LSB unsigned words (with the right size) that
        encode the payload.
    sample_shape : tuple, optional
        Shape of the samples; e.g., (nchan,).  Default: ().
    bps : int, optional
        Bits per elementary sample.  Default: 2.
    complex_data : bool, optional
        Whether data are complex.  Default: `False`.
    """

    _encoders = {4: encode_4bit,
                 8: encode_8bit}
    _decoders = {4: decode_4bit,
                 8: decode_8bit}
    _dtype_word = np.int8

    _sample_shape_maker_1thread = namedtuple('SampleShape', 'nchan')
    _sample_shape_maker_nthread = namedtuple('SampleShape', 'nthread, nchan')

    @classmethod
    def _sample_shape_maker(cls, *args):
        if len(args) == 1:
            return cls._sample_shape_maker_1thread(*args)
        else:
            return cls._sample_shape_maker_nthread(*args)

    @classmethod
    def fromfile(cls, fh, *, payload_nbytes=1 << 22, sample_shape=(1,), bps=4,
                 complex_data=False):
        """Read payloads from several threads.

        Parameters
        ----------
        fh : filehandle or tuple of tuple of filehandle
            Handles to the sets of files from which data are read.  The outer
            tuple holds distinct threads, while the inner ones holds parts of
            those threads.  Typically, these are the two polarisations and the
            two parts of each in which phased baseband data are stored.
        payload_nbytes : int, optional
            Number of bytes to read from each part.  Default: 2**22 (4 MB)
        sample_shape : tuple
            Shape of the samples (e.g., (nchan,) or (nthread, nchan)).
            Default: (1,).
        bps : int, optional
            Bits per elementary sample.  Default: 4.
        complex_data : bool, optional
            Whether data are complex.  Default: `False`.
        """
        if hasattr(fh, 'read'):
            return super().fromfile(fh, payload_nbytes=payload_nbytes,
                                    sample_shape=sample_shape, bps=bps,
                                    complex_data=complex_data)

        payloads = [[super(GSBPayload,
                           cls).fromfile(fh1, payload_nbytes=payload_nbytes,
                                         sample_shape=sample_shape[1:],
                                         bps=bps, complex_data=complex_data)
                     for fh1 in fh_set] for fh_set in fh]
        sample_nbytes, extra = divmod(payloads[0][0]._bpfs, 8)
        assert extra == 0, 'Full samples do not fit in integer number of bytes'
        words = np.empty((len(payloads[0]),
                          payloads[0][0].words.size // sample_nbytes,
                          sample_shape[0],
                          sample_nbytes), dtype=cls._dtype_word)
        for payload_set, thread in zip(payloads, words.transpose(2, 0, 1, 3)):
            for payload, part in zip(payload_set, thread):
                part[:] = payload.words.reshape(-1, sample_nbytes)

        return cls(words.ravel(), sample_shape=sample_shape, bps=bps,
                   complex_data=complex_data)

    def tofile(self, fh):
        try:
            fh.write(self.words.tobytes())
        except AttributeError:
            nthread = len(fh)
            assert nthread == self.sample_shape[0]

            words = self.words.reshape(len(fh[0]), -1, nthread,
                                       self._bpfs // nthread // 8)
            for fh_set, thread in zip(fh, words.transpose(2, 0, 1, 3)):
                for fh, part in zip(fh_set, thread):
                    fh.write(part.tobytes())
